#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
LOG = logging.getLogger(__name__)


class CronTrigger(resource.Resource):
    """A resource implements Mistral cron trigger.

    Cron trigger is an object allowing to run workflow on a schedule. User
    specifies what workflow with what input needs to be run and also specifies
    how often it should be run. Pattern property is used to describe the
    frequency of workflow execution.
    """

    support_status = support.SupportStatus(version='5.0.0')

    PROPERTIES = (
        NAME, PATTERN, WORKFLOW, FIRST_TIME, COUNT
    ) = (
        'name', 'pattern', 'workflow', 'first_time', 'count'
    )

    _WORKFLOW_KEYS = (
        WORKFLOW_NAME, WORKFLOW_INPUT
    ) = (
        'name', 'input'
    )

    ATTRIBUTES = (
        NEXT_EXECUTION_TIME, REMAINING_EXECUTIONS
    ) = (
        'next_execution_time', 'remaining_executions'
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of the cron trigger.')
        ),
        PATTERN: properties.Schema(
            properties.Schema.STRING,
            _('Cron expression.'),
            constraints=[
                constraints.CustomConstraint(
                    'cron_expression')
            ]
        ),
        WORKFLOW: properties.Schema(
            properties.Schema.MAP,
            _('Workflow to execute.'),
            required=True,
            schema={
                WORKFLOW_NAME: properties.Schema(
                    properties.Schema.STRING,
                    _('Name or ID of the workflow.'),
                    required=True,
                    constraints=[
                        constraints.CustomConstraint('mistral.workflow')
                    ]
                ),
                WORKFLOW_INPUT: properties.Schema(
                    properties.Schema.MAP,
                    _('Input values for the workflow.')
                )
            }
        ),
        FIRST_TIME: properties.Schema(
            properties.Schema.STRING,
            _('Time of the first execution in format "YYYY-MM-DD HH:MM".')
        ),
        COUNT: properties.Schema(
            properties.Schema.INTEGER,
            _('Remaining executions.')
        )
    }

    attributes_schema = {
        NEXT_EXECUTION_TIME: attributes.Schema(
            _('Time of the next execution in format "YYYY-MM-DD HH:MM:SS".'),
            type=attributes.Schema.STRING
        ),
        REMAINING_EXECUTIONS: attributes.Schema(
            _('Number of remaining executions.'),
            type=attributes.Schema.INTEGER
        )
    }

    default_client_name = 'mistral'

    entity = 'cron_triggers'

    def validate(self):
        super(CronTrigger, self).validate()
        if not (self.properties[self.PATTERN]
                or self.properties[self.FIRST_TIME]):
            raise exception.PropertyUnspecifiedError(self.PATTERN,
                                                     self.FIRST_TIME)

    def _cron_trigger_name(self):
        return self.properties.get(self.NAME) or self.physical_resource_name()

    def handle_create(self):
        workflow = self.properties.get(self.WORKFLOW)
        name = self._cron_trigger_name()
        identifier = workflow[self.WORKFLOW_NAME]

        args = {
            'pattern': self.properties.get(self.PATTERN),
            'workflow_input': workflow.get(self.WORKFLOW_INPUT),
            'first_time': self.properties.get(self.FIRST_TIME),
            'count': self.properties.get(self.COUNT)
        }

        cron_trigger = self.client().cron_triggers.create(name, identifier,
                                                          **args)
        self.resource_id_set(cron_trigger.name)

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        trigger = self.client().cron_triggers.get(self.resource_id)
        if name == self.NEXT_EXECUTION_TIME:
            return trigger.next_execution_time
        elif name == self.REMAINING_EXECUTIONS:
            return trigger.remaining_executions

    def get_live_state(self, resource_properties):
        # Currently mistral just deletes cron trigger that was executed
        # (i.e. remaining execution is reached zero). In this case we can't
        # found the cron trigger by mistral api. Suppose that live state of
        # cron trigger is equal to the state stored in heat, otherwise we may
        # go through undesirable update-replace. This behaviour might be
        # changed after
        # https://blueprints.launchpad.net/mistral/+spec/mistral-cron-trigger-life-cycle
        # will be merged.
        LOG.warning("get_live_state isn't implemented for this type of "
                    "resource due to specific behaviour of cron trigger "
                    "in mistral.")
        return {}


def resource_mapping():
    return {
        'OS::Mistral::CronTrigger': CronTrigger,
    }



# Copyright 2014 Intel Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""UserCreds object."""

from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from heat.db.sqlalchemy import api as db_api
from heat.objects import base as heat_base


@base.VersionedObjectRegistry.register
class UserCreds(
        heat_base.HeatObject,
        base.VersionedObjectDictCompat,
        base.ComparableVersionedObject,
):
    fields = {
        'id': fields.StringField(),
        'created_at': fields.DateTimeField(read_only=True),
        'updated_at': fields.DateTimeField(nullable=True),
        'username': fields.StringField(nullable=True),
        'password': fields.StringField(nullable=True),
        'tenant': fields.StringField(nullable=True),
        'tenant_id': fields.StringField(nullable=True),
        'trustor_user_id': fields.StringField(nullable=True),
        'trust_id': fields.StringField(nullable=True),
        'region_name': fields.StringField(nullable=True),
        'auth_url': fields.StringField(nullable=True),
        'decrypt_method': fields.StringField(nullable=True)
    }

    @staticmethod
    def _from_db_object(ucreds, db_ucreds, context=None):
        if db_ucreds is None:
            return db_ucreds
        ucreds._context = context
        for field in ucreds.fields:
            # TODO(Shao HE Feng), now the DB layer delete the decrypt_method
            # field, just skip it here. and will add an encrypted_field later.
            if field == "decrypt_method":
                continue
            ucreds[field] = db_ucreds[field]
        ucreds.obj_reset_changes()
        return ucreds

    @classmethod
    def create(cls, context):
        user_creds_db = db_api.user_creds_create(context)
        return cls._from_db_object(cls(), user_creds_db)

    @classmethod
    def delete(cls, context, user_creds_id):
        db_api.user_creds_delete(context, user_creds_id)

    @classmethod
    def get_by_id(cls, context, user_creds_id):
        user_creds_db = db_api.user_creds_get(context, user_creds_id)
        user_creds = cls._from_db_object(cls(), user_creds_db)
        return user_creds



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from heat.common.i18n import _
from heat.engine.resources.openstack.heat import none_resource
from heat.engine import support

LOG = logging.getLogger(__name__)


class Restarter(none_resource.NoneResource):

    support_status = support.SupportStatus(
        status=support.HIDDEN,
        version='10.0.0',
        message=_('The HARestarter resource type has been removed. Existing '
                  'stacks containing HARestarter resources can still be '
                  'used, but the HARestarter resource will be a placeholder '
                  'that does nothing.'),
        previous_status=support.SupportStatus(
            status=support.DEPRECATED,
            message=_('The HARestarter resource type is deprecated and will '
                      'be removed in a future release of Heat, once it has '
                      'support for auto-healing any type of resource. Note '
                      'that HARestarter does *not* actually restart '
                      'servers - it deletes and then recreates them. It also '
                      'does the same to all dependent resources, and may '
                      'therefore exhibit unexpected and undesirable '
                      'behaviour. Instead, use the mark-unhealthy API to '
                      'mark a resource as needing replacement, and then a '
                      'stack update to perform the replacement while '
                      'respecting  the dependencies and not deleting them '
                      'unnecessarily.'),
            version='2015.1'))


def resource_mapping():
    return {
        'OS::Heat::HARestarter': Restarter,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

import re
import yaml

from heat.common import config
from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import neutron
from heat.tests import common
from heat.tests import utils


class JsonToYamlTest(common.HeatTestCase):

    def setUp(self):
        super(JsonToYamlTest, self).setUp()
        self.expected_test_count = 2
        self.longMessage = True
        self.maxDiff = None

    def test_convert_all_templates(self):
        path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            'templates')

        template_test_count = 0
        for (json_str,
             yml_str) in self.convert_all_json_to_yaml(path):

            self.compare_json_vs_yaml(json_str, yml_str)
            template_test_count += 1
            if template_test_count >= self.expected_test_count:
                break

        self.assertGreaterEqual(
            template_test_count, self.expected_test_count,
            'Expected at least %d templates to be tested, not %d' %
            (self.expected_test_count, template_test_count))

    def compare_json_vs_yaml(self, json_str, yml_str):
        yml = template_format.parse(yml_str)

        self.assertEqual(u'2012-12-12', yml[u'HeatTemplateFormatVersion'])
        self.assertNotIn(u'AWSTemplateFormatVersion', yml)
        del(yml[u'HeatTemplateFormatVersion'])

        jsn = template_format.parse(json_str)

        if u'AWSTemplateFormatVersion' in jsn:
            del(jsn[u'AWSTemplateFormatVersion'])

        self.assertEqual(yml, jsn)

    def convert_all_json_to_yaml(self, dirpath):
        for path in os.listdir(dirpath):
            if not path.endswith('.template') and not path.endswith('.json'):
                continue
            with open(os.path.join(dirpath, path), 'r') as f:
                json_str = f.read()

            yml_str = template_format.convert_json_to_yaml(json_str)
            yield (json_str, yml_str)

    def test_integer_only_keys_get_translated_correctly(self):
        path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            'templates/WordPress_Single_Instance.template')
        with open(path, 'r') as f:
            json_str = f.read()
            yml_str = template_format.convert_json_to_yaml(json_str)
            match = re.search(r'[\s,{]\d+\s*:', yml_str)
            # Check that there are no matches of integer-only keys
            # lacking explicit quotes
            self.assertIsNone(match)


class YamlMinimalTest(common.HeatTestCase):

    def _parse_template(self, tmpl_str, msg_str):
        parse_ex = self.assertRaises(ValueError,
                                     template_format.parse,
                                     tmpl_str)
        self.assertIn(msg_str, str(parse_ex))

    def test_long_yaml(self):
        template = {'HeatTemplateFormatVersion': '2012-12-12'}
        config.cfg.CONF.set_override('max_template_size', 10)
        template['Resources'] = ['a'] * int(
            config.cfg.CONF.max_template_size / 3)
        limit = config.cfg.CONF.max_template_size
        long_yaml = yaml.safe_dump(template)
        self.assertGreater(len(long_yaml), limit)
        ex = self.assertRaises(exception.RequestLimitExceeded,
                               template_format.parse, long_yaml)
        msg = ('Request limit exceeded: Template size (%(actual_len)s '
               'bytes) exceeds maximum allowed size (%(limit)s bytes).') % {
                   'actual_len': len(str(long_yaml)),
                   'limit': config.cfg.CONF.max_template_size}
        self.assertEqual(msg, str(ex))

    def test_parse_no_version_format(self):
        yaml = ''
        self._parse_template(yaml, 'Template format version not found')
        yaml2 = '''Parameters: {}
Mappings: {}
Resources: {}
Outputs: {}
'''
        self._parse_template(yaml2, 'Template format version not found')

    def test_parse_string_template(self):
        tmpl_str = 'just string'
        msg = 'The template is not a JSON object or YAML mapping.'
        self._parse_template(tmpl_str, msg)

    def test_parse_invalid_yaml_and_json_template(self):
        tmpl_str = '{test'
        msg = 'line 1, column 1'
        self._parse_template(tmpl_str, msg)

    def test_parse_json_document(self):
        tmpl_str = '["foo" , "bar"]'
        msg = 'The template is not a JSON object or YAML mapping.'
        self._parse_template(tmpl_str, msg)

    def test_parse_empty_json_template(self):
        tmpl_str = '{}'
        msg = 'Template format version not found'
        self._parse_template(tmpl_str, msg)

    def test_parse_yaml_template(self):
        tmpl_str = 'heat_template_version: 2013-05-23'
        expected = {'heat_template_version': '2013-05-23'}
        self.assertEqual(expected, template_format.parse(tmpl_str))


class YamlParseExceptions(common.HeatTestCase):

    scenarios = [
        ('scanner', dict(raised_exception=yaml.scanner.ScannerError())),
        ('parser', dict(raised_exception=yaml.parser.ParserError())),
        ('reader',
         dict(raised_exception=yaml.reader.ReaderError(
             '', 42, 'x'.encode('latin-1'), '', ''))),
    ]

    def test_parse_to_value_exception(self):
        text = 'not important'

        with mock.patch.object(yaml, 'load') as yaml_loader:
            yaml_loader.side_effect = self.raised_exception

            err = self.assertRaises(ValueError,
                                    template_format.parse, text,
                                    'file://test.yaml')

            self.assertIn('Error parsing template file://test.yaml',
                          str(err))


class JsonYamlResolvedCompareTest(common.HeatTestCase):

    def setUp(self):
        super(JsonYamlResolvedCompareTest, self).setUp()
        self.longMessage = True
        self.maxDiff = None

    def load_template(self, file_name):
        filepath = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                'templates', file_name)
        f = open(filepath)
        t = template_format.parse(f.read())
        f.close()
        return t

    def compare_stacks(self, json_file, yaml_file, parameters):
        t1 = self.load_template(json_file)
        t2 = self.load_template(yaml_file)
        del(t1[u'AWSTemplateFormatVersion'])
        t1[u'HeatTemplateFormatVersion'] = t2[u'HeatTemplateFormatVersion']
        stack1 = utils.parse_stack(t1, parameters)
        stack2 = utils.parse_stack(t2, parameters)

        # compare resources separately so that resolved static data
        # is compared
        t1nr = dict(stack1.t.t)
        del(t1nr['Resources'])

        t2nr = dict(stack2.t.t)
        del(t2nr['Resources'])
        self.assertEqual(t1nr, t2nr)

        self.assertEqual(set(stack1), set(stack2))
        for key in stack1:
            self.assertEqual(stack1[key].t, stack2[key].t)

    def test_neutron_resolved(self):
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)
        self.compare_stacks('Neutron.template', 'Neutron.yaml', {})

    def test_wordpress_resolved(self):
        self.compare_stacks('WordPress_Single_Instance.template',
                            'WordPress_Single_Instance.yaml',
                            {'KeyName': 'test'})




#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine.resources.openstack.neutron import router
from heat.engine import support


class ExtraRoute(neutron.NeutronResource):
    """Resource for specifying extra routes for Neutron router.

    Resource allows to specify nexthop IP and destination network for router.
    """

    required_service_extension = 'extraroute'

    support_status = support.SupportStatus(
        status=support.UNSUPPORTED,
        message=_('Use this resource at your own risk.'))

    PROPERTIES = (
        ROUTER_ID, DESTINATION, NEXTHOP,
    ) = (
        'router_id', 'destination', 'nexthop',
    )

    properties_schema = {
        ROUTER_ID: properties.Schema(
            properties.Schema.STRING,
            description=_('The router id.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('neutron.router')
            ]
        ),
        DESTINATION: properties.Schema(
            properties.Schema.STRING,
            description=_('Network in CIDR notation.'),
            required=True),
        NEXTHOP: properties.Schema(
            properties.Schema.STRING,
            description=_('Nexthop IP address.'),
            required=True)
    }

    def add_dependencies(self, deps):
        super(ExtraRoute, self).add_dependencies(deps)
        for resource in self.stack.values():
            # depend on any RouterInterface in this template with the same
            # router_id as this router_id
            if resource.has_interface('OS::Neutron::RouterInterface'):
                try:
                    router_id = self.properties[self.ROUTER_ID]
                    dep_router_id = resource.properties.get(
                        router.RouterInterface.ROUTER)
                except (ValueError, TypeError):
                    # Properties errors will be caught later in validation,
                    # where we can report them in their proper context.
                    continue
                if dep_router_id == router_id:
                    deps += (self, resource)
            # depend on any RouterGateway in this template with the same
            # router_id as this router_id
            elif resource.has_interface('OS::Neutron::RouterGateway'):
                try:
                    router_id = self.properties[self.ROUTER_ID]
                    dep_router_id = resource.properties.get(
                        router.RouterGateway.ROUTER_ID)
                except (ValueError, TypeError):
                    # Properties errors will be caught later in validation,
                    # where we can report them in their proper context.
                    continue
                if dep_router_id == router_id:
                    deps += (self, resource)

    def handle_create(self):
        router_id = self.properties.get(self.ROUTER_ID)
        routes = self.client().show_router(
            router_id).get('router').get('routes')
        if not routes:
            routes = []
        new_route = {'destination': self.properties[self.DESTINATION],
                     'nexthop': self.properties[self.NEXTHOP]}
        if new_route in routes:
            msg = _('Route duplicates an existing route.')
            raise exception.Error(msg)
        routes.append(new_route.copy())
        self.client().update_router(router_id,
                                    {'router': {'routes': routes}})
        new_route['router_id'] = router_id
        self.resource_id_set(
            '%(router_id)s:%(destination)s:%(nexthop)s' % new_route)

    def handle_delete(self):
        if not self.resource_id:
            return
        router_id = self.properties[self.ROUTER_ID]
        with self.client_plugin().ignore_not_found:
            routes = self.client().show_router(
                router_id).get('router').get('routes', [])
            try:
                routes.remove(
                    {'destination': self.properties[self.DESTINATION],
                     'nexthop': self.properties[self.NEXTHOP]}
                )
            except ValueError:
                return
            self.client().update_router(router_id,
                                        {'router': {'routes': routes}})


def resource_mapping():
    return {
        'OS::Neutron::ExtraRoute': ExtraRoute,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from heat.common import exception
from heat.engine import properties_group as pg
from heat.tests import common


class TestSchemaSimpleValidation(common.HeatTestCase):

    scenarios = [
        ('correct schema', dict(
            schema={pg.AND: [['a'], ['b']]},
            message=None,
        )),
        ('invalid type schema', dict(
            schema=[{pg.OR: [['a'], ['b']]}],
            message="Properties group schema incorrectly specified. "
                    "Schema should be a mapping, "
                    "found %s instead." % list,
        )),
        ('invalid type subschema', dict(
            schema={pg.OR: [['a'], ['b'], [{pg.XOR: [['c'], ['d']]}]]},
            message='Properties group schema incorrectly specified. List '
                    'items should be properties list-type names with format '
                    '"[prop, prop_child, prop_sub_child, ...]" or nested '
                    'properties group schemas.',
        )),
        ('several keys schema', dict(
            schema={pg.OR: [['a'], ['b']],
                    pg.XOR: [['v', 'g']]},
            message='Properties group schema incorrectly specified. Schema '
                    'should be one-key dict.',
        )),
        ('several keys subschema', dict(
            schema={pg.OR: [['a'], ['b'], {pg.XOR: [['c']], pg.OR: ['d']}]},
            message='Properties group schema incorrectly specified. '
                    'Schema should be one-key dict.',
        )),
        ('invalid key schema', dict(
            schema={'NOT KEY': [['a'], ['b']]},
            message='Properties group schema incorrectly specified. '
                    'Properties group schema key should be one of the '
                    'operators: AND, OR, XOR.',
        )),
        ('invalid key subschema', dict(
            schema={pg.AND: [['a'], {'NOT KEY': [['b']]}]},
            message='Properties group schema incorrectly specified. '
                    'Properties group schema key should be one of the '
                    'operators: AND, OR, XOR.',
        )),
        ('invalid value type schema', dict(
            schema={pg.OR: 'a'},
            message="Properties group schema incorrectly specified. "
                    "Schemas' values should be lists of properties names "
                    "or nested schemas.",
        )),
        ('invalid value type subschema', dict(
            schema={pg.OR: [{pg.XOR: 'a'}]},
            message="Properties group schema incorrectly specified. "
                    "Schemas' values should be lists of properties names "
                    "or nested schemas.",
        )),
        ('invalid prop name schema', dict(
            schema={pg.OR: ['a', 'b']},
            message='Properties group schema incorrectly specified. List '
                    'items should be properties list-type names with format '
                    '"[prop, prop_child, prop_sub_child, ...]" or nested '
                    'properties group schemas.',
        )),
    ]

    def test_properties_group_schema_validate(self):
        if self.message is not None:
            ex = self.assertRaises(exception.InvalidSchemaError,
                                   pg.PropertiesGroup, self.schema)
            self.assertEqual(self.message, str(ex))
        else:
            self.assertIsInstance(pg.PropertiesGroup(self.schema),
                                  pg.PropertiesGroup)



# Copyright 2015 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from unittest import mock

from oslo_config import cfg

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import senlin
from heat.engine.resources.openstack.senlin import cluster as sc
from heat.engine import scheduler
from heat.engine import template
from heat.tests import common
from heat.tests import utils
from openstack import exceptions


cluster_stack_template = """
heat_template_version: 2016-04-08
description: Senlin Cluster Template
resources:
  senlin-cluster:
    type: OS::Senlin::Cluster
    properties:
      name: SenlinCluster
      profile: fake_profile
      policies:
        - policy: fake_policy
          enabled: true
      min_size: 0
      max_size: -1
      desired_capacity: 1
      timeout: 3600
      metadata:
        foo: bar
"""


class FakeCluster(object):
    def __init__(self, id='some_id', status='ACTIVE'):
        self.status = status
        self.status_reason = 'Unknown'
        self.id = id
        self.name = "SenlinCluster"
        self.metadata = {}
        self.nodes = ['node1']
        self.desired_capacity = 1
        self.metadata = {'foo': 'bar'}
        self.timeout = 3600
        self.max_size = -1
        self.min_size = 0
        self.location = 'actions/fake-action'
        self.profile_name = 'fake_profile'
        self.profile_id = 'fake_profile_id'

    def to_dict(self):
        return {
            'id': self.id,
            'status': self.status,
            'status_reason': self.status_reason,
            'name': self.name,
            'metadata': self.metadata,
            'timeout': self.timeout,
            'desired_capacity': self.desired_capacity,
            'max_size': self.max_size,
            'min_size': self.min_size,
            'nodes': self.nodes,
            'profile_name': self.profile_name,
            'profile_id': self.profile_id
        }


class SenlinClusterTest(common.HeatTestCase):
    def setUp(self):
        super(SenlinClusterTest, self).setUp()
        self.senlin_mock = mock.MagicMock()
        self.senlin_mock.get_profile.return_value = mock.Mock(
            id='fake_profile_id'
        )
        self.patchobject(sc.Cluster, 'client', return_value=self.senlin_mock)
        self.patchobject(senlin.SenlinClientPlugin, 'client',
                         return_value=self.senlin_mock)
        self.patchobject(senlin.ProfileConstraint, 'validate',
                         return_value=True)
        self.patchobject(senlin.PolicyConstraint, 'validate',
                         return_value=True)
        self.fake_cl = FakeCluster()
        self.t = template_format.parse(cluster_stack_template)

    def _init_cluster(self, template):
        self.stack = utils.parse_stack(template)
        cluster = self.stack['senlin-cluster']
        return cluster

    def _create_cluster(self, template):
        cluster = self._init_cluster(template)
        self.senlin_mock.create_cluster.return_value = self.fake_cl
        self.senlin_mock.get_cluster.return_value = self.fake_cl
        self.senlin_mock.get_action.return_value = mock.Mock(
            status='SUCCEEDED')
        self.senlin_mock.get_policy.return_value = mock.Mock(
            id='fake_policy_id'
        )
        self.senlin_mock.cluster_policies.return_value = [
            {'policy_id': 'fake_policy_id', 'enabled': True}
        ]
        scheduler.TaskRunner(cluster.create)()
        self.assertEqual((cluster.CREATE, cluster.COMPLETE),
                         cluster.state)
        self.assertEqual(self.fake_cl.id, cluster.resource_id)
        self.assertEqual(1, self.senlin_mock.get_action.call_count)
        self.assertEqual(1, self.senlin_mock.get_cluster.call_count)
        return cluster

    def test_cluster_create_success(self):
        self._create_cluster(self.t)
        create_cluster_kwargs = {
            'name': 'SenlinCluster',
            'profile_id': 'fake_profile_id',
            'desired_capacity': 1,
            'min_size': 0,
            'max_size': -1,
            'metadata': {'foo': 'bar'},
            'timeout': 3600,
        }
        attach_policy_kwargs = {
            'cluster': self.fake_cl.id,
            'policy': 'fake_policy_id',
            'enabled': True
        }
        self.senlin_mock.create_cluster.assert_called_once_with(
            **create_cluster_kwargs)
        self.senlin_mock.cluster_attach_policy.assert_called_once_with(
            **attach_policy_kwargs)

    def test_cluster_create_error(self):
        cfg.CONF.set_override('action_retry_limit', 0)
        cluster = self._init_cluster(self.t)
        self.senlin_mock.create_cluster.return_value = self.fake_cl
        mock_cluster = mock.MagicMock()
        mock_cluster.status = 'ERROR'
        mock_cluster.status_reason = 'oops'
        self.senlin_mock.get_policy.return_value = mock.Mock(
            id='fake_policy_id'
        )
        self.senlin_mock.get_cluster.return_value = mock_cluster
        create_task = scheduler.TaskRunner(cluster.create)
        ex = self.assertRaises(exception.ResourceFailure, create_task)
        expected = ('ResourceInError: resources.senlin-cluster: '
                    'Went to status ERROR due to "oops"')
        self.assertEqual(expected, str(ex))

    def test_cluster_delete_success(self):
        cluster = self._create_cluster(self.t)
        self.senlin_mock.get_cluster.side_effect = [
            exceptions.ResourceNotFound('SenlinCluster'),
        ]
        scheduler.TaskRunner(cluster.delete)()
        self.senlin_mock.delete_cluster.assert_called_once_with(
            cluster.resource_id)

    def test_cluster_delete_error(self):
        cluster = self._create_cluster(self.t)
        self.senlin_mock.get_cluster.side_effect = exception.Error('oops')
        delete_task = scheduler.TaskRunner(cluster.delete)
        ex = self.assertRaises(exception.ResourceFailure, delete_task)
        expected = 'Error: resources.senlin-cluster: oops'
        self.assertEqual(expected, str(ex))

    def test_cluster_update_profile(self):
        cluster = self._create_cluster(self.t)
        # Mock translate rules
        self.senlin_mock.get_profile.side_effect = [
            mock.Mock(id='new_profile_id'),
            mock.Mock(id='fake_profile_id'),
            mock.Mock(id='new_profile_id'),
        ]
        new_t = copy.deepcopy(self.t)
        props = new_t['resources']['senlin-cluster']['properties']
        props['profile'] = 'new_profile'
        props['name'] = 'new_name'
        rsrc_defns = template.Template(new_t).resource_definitions(self.stack)
        new_cluster = rsrc_defns['senlin-cluster']
        self.senlin_mock.update_cluster.return_value = mock.Mock(
            location='/actions/fake-action')
        self.senlin_mock.get_action.return_value = mock.Mock(
            status='SUCCEEDED')
        scheduler.TaskRunner(cluster.update, new_cluster)()
        self.assertEqual((cluster.UPDATE, cluster.COMPLETE), cluster.state)
        cluster_update_kwargs = {
            'profile_id': 'new_profile_id',
            'name': 'new_name'
        }
        self.senlin_mock.update_cluster.assert_called_once_with(
            cluster=self.fake_cl, **cluster_update_kwargs)
        self.assertEqual(2, self.senlin_mock.get_action.call_count)

    def test_cluster_update_desire_capacity(self):
        cluster = self._create_cluster(self.t)
        new_t = copy.deepcopy(self.t)
        props = new_t['resources']['senlin-cluster']['properties']
        props['desired_capacity'] = 10
        rsrc_defns = template.Template(new_t).resource_definitions(self.stack)
        new_cluster = rsrc_defns['senlin-cluster']
        self.senlin_mock.cluster_resize.return_value = {
            'action': 'fake-action'}
        self.senlin_mock.get_action.return_value = mock.Mock(
            status='SUCCEEDED')
        scheduler.TaskRunner(cluster.update, new_cluster)()
        self.assertEqual((cluster.UPDATE, cluster.COMPLETE), cluster.state)
        cluster_resize_kwargs = {
            'adjustment_type': 'EXACT_CAPACITY',
            'number': 10
        }
        self.senlin_mock.cluster_resize.assert_called_once_with(
            cluster=cluster.resource_id, **cluster_resize_kwargs)
        self.assertEqual(2, self.senlin_mock.get_action.call_count)

    def test_cluster_update_policy_add_remove(self):
        cluster = self._create_cluster(self.t)
        # Mock translate rules
        self.senlin_mock.get_policy.side_effect = [
            mock.Mock(id='new_policy_id'),
            mock.Mock(id='fake_policy_id'),
            mock.Mock(id='new_policy_id'),
        ]
        new_t = copy.deepcopy(self.t)
        props = new_t['resources']['senlin-cluster']['properties']
        props['policies'] = [{'policy': 'new_policy'}]
        rsrc_defns = template.Template(new_t).resource_definitions(self.stack)
        new_cluster = rsrc_defns['senlin-cluster']
        self.senlin_mock.cluster_detach_policy.return_value = {
            'action': 'fake-action'}
        self.senlin_mock.cluster_attach_policy.return_value = {
            'action': 'fake-action'}
        self.senlin_mock.get_action.return_value = mock.Mock(
            status='SUCCEEDED')
        scheduler.TaskRunner(cluster.update, new_cluster)()
        self.assertEqual((cluster.UPDATE, cluster.COMPLETE), cluster.state)
        detach_policy_kwargs = {
            'policy': 'fake_policy_id',
            'cluster': cluster.resource_id,
            'enabled': True,
        }
        self.assertEqual(2,
                         self.senlin_mock.cluster_attach_policy.call_count)
        self.senlin_mock.cluster_detach_policy.assert_called_once_with(
            **detach_policy_kwargs)
        self.assertEqual(0, self.senlin_mock.cluster_update_policy.call_count)
        self.assertEqual(3, self.senlin_mock.get_action.call_count)

    def test_cluster_update_policy_exists(self):
        cluster = self._create_cluster(self.t)
        new_t = copy.deepcopy(self.t)
        props = new_t['resources']['senlin-cluster']['properties']
        props['policies'] = [{'policy': 'fake_policy', 'enabled': False}]
        rsrc_defns = template.Template(new_t).resource_definitions(self.stack)
        new_cluster = rsrc_defns['senlin-cluster']
        self.senlin_mock.cluster_update_policy.return_value = {
            'action': 'fake-action'}
        self.senlin_mock.get_action.return_value = mock.Mock(
            status='SUCCEEDED')
        scheduler.TaskRunner(cluster.update, new_cluster)()
        self.assertEqual((cluster.UPDATE, cluster.COMPLETE), cluster.state)
        update_policy_kwargs = {
            'policy': 'fake_policy_id',
            'cluster': cluster.resource_id,
            'enabled': False,
        }
        self.senlin_mock.cluster_update_policy.assert_called_once_with(
            **update_policy_kwargs)
        self.assertEqual(1, self.senlin_mock.cluster_attach_policy.call_count)
        self.assertEqual(0, self.senlin_mock.cluster_detach_policy.call_count)

    def test_cluster_update_failed(self):
        cluster = self._create_cluster(self.t)
        new_t = copy.deepcopy(self.t)
        props = new_t['resources']['senlin-cluster']['properties']
        props['desired_capacity'] = 3
        rsrc_defns = template.Template(new_t).resource_definitions(self.stack)
        update_snippet = rsrc_defns['senlin-cluster']
        self.senlin_mock.cluster_resize.return_value = {
            'action': 'fake-action'}
        self.senlin_mock.get_action.return_value = mock.Mock(
            status='FAILED', status_reason='Unknown')
        exc = self.assertRaises(
            exception.ResourceFailure,
            scheduler.TaskRunner(cluster.update, update_snippet))
        self.assertEqual('ResourceInError: resources.senlin-cluster: '
                         'Went to status FAILED due to "Unknown"',
                         str(exc))

    def test_cluster_get_attr_collect(self):
        cluster = self._create_cluster(self.t)
        self.senlin_mock.collect_cluster_attrs.return_value = [
            mock.Mock(attr_value='ip1')]
        attr_path1 = ['details.addresses.private[0].addr']
        self.assertEqual(
            ['ip1'], cluster.get_attribute(cluster.ATTR_COLLECT, *attr_path1))
        attr_path2 = ['details.addresses.private[0].addr', 0]
        self.assertEqual(
            'ip1', cluster.get_attribute(cluster.ATTR_COLLECT, *attr_path2))
        self.senlin_mock.collect_cluster_attrs.assert_called_with(
            cluster.resource_id, attr_path2[0])

    def test_cluster_resolve_attribute(self):
        excepted_show = {
            'id': 'some_id',
            'status': 'ACTIVE',
            'status_reason': 'Unknown',
            'name': 'SenlinCluster',
            'metadata': {'foo': 'bar'},
            'timeout': 3600,
            'desired_capacity': 1,
            'max_size': -1,
            'min_size': 0,
            'nodes': ['node1'],
            'profile_name': 'fake_profile',
            'profile_id': 'fake_profile_id',
            'policies': [{'policy_id': 'fake_policy_id', 'enabled': True}]
        }
        cluster = self._create_cluster(self.t)
        self.assertEqual(self.fake_cl.desired_capacity,
                         cluster._resolve_attribute('desired_capacity'))
        self.assertEqual(['node1'],
                         cluster._resolve_attribute('nodes'))
        self.assertEqual(excepted_show,
                         cluster._show_resource())

    def test_cluster_get_live_state(self):
        expected_reality = {
            'name': 'SenlinCluster',
            'metadata': {'foo': 'bar'},
            'timeout': 3600,
            'desired_capacity': 1,
            'max_size': -1,
            'min_size': 0,
            'profile': 'fake_profile_id',
            'policies': [{'policy': 'fake_policy_id', 'enabled': True}]
        }
        cluster = self._create_cluster(self.t)
        self.senlin_mock.get_cluster.return_value = self.fake_cl

        reality = cluster.get_live_state(cluster.properties)

        self.assertEqual(expected_reality, reality)


class TestSenlinClusterValidation(common.HeatTestCase):
    def setUp(self):
        super(TestSenlinClusterValidation, self).setUp()
        self.t = template_format.parse(cluster_stack_template)

    def test_invalid_min_max_size(self):
        self.t['resources']['senlin-cluster']['properties']['min_size'] = 2
        self.t['resources']['senlin-cluster']['properties']['max_size'] = 1
        stack = utils.parse_stack(self.t)
        ex = self.assertRaises(exception.StackValidationFailed,
                               stack['senlin-cluster'].validate)
        self.assertEqual('min_size can not be greater than max_size',
                         str(ex))

    def test_invalid_desired_capacity(self):
        self.t['resources']['senlin-cluster']['properties']['min_size'] = 1
        self.t['resources']['senlin-cluster']['properties']['max_size'] = 2
        self.t['resources']['senlin-cluster']['properties'][
            'desired_capacity'] = 3
        stack = utils.parse_stack(self.t)
        ex = self.assertRaises(exception.StackValidationFailed,
                               stack['senlin-cluster'].validate)
        self.assertEqual(
            'desired_capacity must be between min_size and max_size',
            str(ex)
        )



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from urllib import parse

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import properties
from heat.engine import resource
from heat.engine import support

LOG = logging.getLogger(__name__)


class SwiftContainer(resource.Resource):
    """A resource for managing Swift containers.

    A container defines a namespace for objects. An object with the same
    name in two different containers represents two different objects.
    """
    PROPERTIES = (
        NAME, X_CONTAINER_READ, X_CONTAINER_WRITE, X_CONTAINER_META,
        X_ACCOUNT_META, PURGE_ON_DELETE,
    ) = (
        'name', 'X-Container-Read', 'X-Container-Write', 'X-Container-Meta',
        'X-Account-Meta', 'PurgeOnDelete',
    )

    ATTRIBUTES = (
        DOMAIN_NAME, WEBSITE_URL, ROOT_URL, OBJECT_COUNT, BYTES_USED,
        HEAD_CONTAINER,
    ) = (
        'DomainName', 'WebsiteURL', 'RootURL', 'ObjectCount', 'BytesUsed',
        'HeadContainer',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name for the container. If not specified, a unique name will '
              'be generated.')
        ),
        X_CONTAINER_READ: properties.Schema(
            properties.Schema.STRING,
            _('Specify the ACL permissions on who can read objects in the '
              'container.')
        ),
        X_CONTAINER_WRITE: properties.Schema(
            properties.Schema.STRING,
            _('Specify the ACL permissions on who can write objects to the '
              'container.')
        ),
        X_CONTAINER_META: properties.Schema(
            properties.Schema.MAP,
            _('A map of user-defined meta data to associate with the '
              'container. Each key in the map will set the header '
              'X-Container-Meta-{key} with the corresponding value.'),
            default={}
        ),
        X_ACCOUNT_META: properties.Schema(
            properties.Schema.MAP,
            _('A map of user-defined meta data to associate with the '
              'account. Each key in the map will set the header '
              'X-Account-Meta-{key} with the corresponding value.'),
            default={}
        ),
        PURGE_ON_DELETE: properties.Schema(
            properties.Schema.BOOLEAN,
            _("If True, delete any objects in the container "
              "when the container is deleted. "
              "Otherwise, deleting a non-empty container "
              "will result in an error."),
            default=False,
            support_status=support.SupportStatus(
                version='2015.1')
        ),
    }

    attributes_schema = {
        DOMAIN_NAME: attributes.Schema(
            _('The host from the container URL.'),
            type=attributes.Schema.STRING
        ),
        WEBSITE_URL: attributes.Schema(
            _('The URL of the container.'),
            type=attributes.Schema.STRING
        ),
        ROOT_URL: attributes.Schema(
            _('The parent URL of the container.'),
            type=attributes.Schema.STRING
        ),
        OBJECT_COUNT: attributes.Schema(
            _('The number of objects stored in the container.'),
            type=attributes.Schema.INTEGER
        ),
        BYTES_USED: attributes.Schema(
            _('The number of bytes stored in the container.'),
            type=attributes.Schema.INTEGER
        ),
        HEAD_CONTAINER: attributes.Schema(
            _('A map containing all headers for the container.'),
            type=attributes.Schema.MAP
        ),
    }

    default_client_name = 'swift'

    def physical_resource_name(self):
        name = self.properties[self.NAME]
        if name:
            return name

        return super(SwiftContainer, self).physical_resource_name()

    @staticmethod
    def _build_meta_headers(obj_type, meta_props):
        """Returns a new dict.

        Each key of new dict is prepended with "X-Container-Meta-".
        """
        if meta_props is None:
            return {}
        return dict(
            ('X-' + obj_type.title() + '-Meta-' + k, v)
            for (k, v) in meta_props.items())

    def handle_create(self):
        """Create a container."""
        container = self.physical_resource_name()

        container_headers = SwiftContainer._build_meta_headers(
            "container", self.properties[self.X_CONTAINER_META])

        account_headers = SwiftContainer._build_meta_headers(
            "account", self.properties[self.X_ACCOUNT_META])

        for key in (self.X_CONTAINER_READ, self.X_CONTAINER_WRITE):
            if self.properties[key] is not None:
                container_headers[key] = self.properties[key]

        LOG.debug('SwiftContainer create container %(container)s with '
                  'container headers %(container_headers)s and '
                  'account headers %(account_headers)s',
                  {'container': container,
                   'account_headers': account_headers,
                   'container_headers': container_headers})

        self.client().put_container(container, container_headers)

        if account_headers:
            self.client().post_account(account_headers)

        self.resource_id_set(container)

    def _get_objects(self):
        try:
            container, objects = self.client().get_container(self.resource_id)
        except Exception as ex:
            self.client_plugin().ignore_not_found(ex)
            return None
        return objects

    def _deleter(self, obj=None):
        """Delete the underlying container or an object inside it."""
        args = [self.resource_id]
        if obj:
            deleter = self.client().delete_object
            args.append(obj['name'])
        else:
            deleter = self.client().delete_container

        with self.client_plugin().ignore_not_found:
            deleter(*args)

    def handle_delete(self):
        if self.resource_id is None:
            return

        objects = self._get_objects()

        if objects:
            if self.properties[self.PURGE_ON_DELETE]:
                self._deleter(objects.pop())  # save first container refresh
            else:
                msg = _("Deleting non-empty container (%(id)s) "
                        "when %(prop)s is False") % {
                            'id': self.resource_id,
                            'prop': self.PURGE_ON_DELETE}
                raise exception.ResourceActionNotSupported(action=msg)
        # objects is either None (container is gone already) or (empty) list
        if objects is not None:
            objects = len(objects)
        return objects

    def check_delete_complete(self, objects):
        if objects is None:  # resource was not created or is gone already
            return True
        if objects:  # integer >=0 from the first invocation
            objs = self._get_objects()
            if objs is None:
                return True  # container is gone already
            if objs:
                self._deleter(objs.pop())
                if objs:  # save one last _get_objects() API call
                    return False

        self._deleter()
        return True

    def handle_check(self):
        self.client().get_container(self.resource_id)

    def get_reference_id(self):
        return str(self.resource_id)

    def _resolve_attribute(self, key):
        parsed = list(parse.urlparse(self.client().url))
        if key == self.DOMAIN_NAME:
            return parsed[1].split(':')[0]
        elif key == self.WEBSITE_URL:
            return '%s://%s%s/%s' % (parsed[0], parsed[1], parsed[2],
                                     self.resource_id)
        elif key == self.ROOT_URL:
            return '%s://%s%s' % (parsed[0], parsed[1], parsed[2])
        elif self.resource_id and key in (
                self.OBJECT_COUNT, self.BYTES_USED, self.HEAD_CONTAINER):
            try:
                headers = self.client().head_container(self.resource_id)
            except Exception as ex:
                if self.client_plugin().is_client_exception(ex):
                    LOG.warning("Head container failed: %s", ex)
                    return None
                raise
            else:
                if key == self.OBJECT_COUNT:
                    return headers['x-container-object-count']
                elif key == self.BYTES_USED:
                    return headers['x-container-bytes-used']
                elif key == self.HEAD_CONTAINER:
                    return headers

    def _show_resource(self):
        return self.client().head_container(self.resource_id)

    def get_live_resource_data(self):
        resource_data = super(SwiftContainer, self).get_live_resource_data()
        account_data = self.client().head_account()
        resource_data['account_data'] = account_data
        return resource_data

    def parse_live_resource_data(self, resource_properties, resource_data):
        swift_reality = {}
        # swift container name can't be updated
        swift_reality.update({self.NAME: resource_properties.get(self.NAME)})
        # PURGE_ON_DELETE property is used only on the heat side and isn't
        # passed to swift, so update it from existing resource properties
        swift_reality.update({self.PURGE_ON_DELETE: resource_properties.get(
            self.PURGE_ON_DELETE)})
        swift_reality.update({self.X_CONTAINER_META: {}})
        swift_reality.update({self.X_ACCOUNT_META: {}})
        for key in [self.X_CONTAINER_READ, self.X_CONTAINER_WRITE]:
            swift_reality.update({key: resource_data.get(key.lower())})
        for key in resource_properties.get(self.X_CONTAINER_META):
            prefixed_key = "%(prefix)s-%(key)s" % {
                'prefix': self.X_CONTAINER_META.lower(),
                'key': key.lower()}
            if prefixed_key in resource_data:
                swift_reality[self.X_CONTAINER_META].update(
                    {key: resource_data[prefixed_key]})
        for key in resource_properties.get(self.X_ACCOUNT_META):
            prefixed_key = "%(prefix)s-%(key)s" % {
                'prefix': self.X_ACCOUNT_META.lower(),
                'key': key.lower()}
            if prefixed_key in resource_data['account_data']:
                swift_reality[self.X_ACCOUNT_META].update(
                    {key: resource_data['account_data'][prefixed_key]})
        return swift_reality


def resource_mapping():
    return {
        'OS::Swift::Container': SwiftContainer,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from oslo_log import log as logging


LOG = logging.getLogger(__name__)


def list_all():
    scenario_dir = os.path.join(os.path.dirname(__file__), '../scenarios')
    if not os.path.isdir(scenario_dir):
        LOG.error('Scenario directory "%s" not found', scenario_dir)
        return

    for root, dirs, files in os.walk(scenario_dir):
        for filename in files:
            name, ext = os.path.splitext(filename)
            if ext == '.py':
                LOG.debug('Found scenario "%s"', name)
                yield name, os.path.join(root, filename)


class Scenario(object):

    def __init__(self, name, path):
        self.name = name

        with open(path) as f:
            source = f.read()

        self.code = compile(source, path, 'exec')
        LOG.debug('Loaded scenario %s', self.name)

    def __call__(self, _event_loop, **global_env):
        LOG.info('*** Beginning scenario "%s"', self.name)

        exec(self.code, global_env, {})
        _event_loop()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import re

from oslo_config import cfg
from oslo_log import log
from oslo_messaging._drivers import common as rpc_common
import webob

import heat.api.middleware.fault as fault
from heat.common import exception as heat_exc
from heat.common.i18n import _
from heat.tests import common


class StackNotFoundChild(heat_exc.EntityNotFound):
    pass


class ErrorWithNewline(webob.exc.HTTPBadRequest):
    pass


class FaultMiddlewareTest(common.HeatTestCase):
    def setUp(self):
        super(FaultMiddlewareTest, self).setUp()
        log.register_options(cfg.CONF)

    def test_disguised_http_exception_with_newline(self):
        wrapper = fault.FaultWrapper(None)
        newline_error = ErrorWithNewline('Error with \n newline')
        msg = wrapper._error(heat_exc.HTTPExceptionDisguise(newline_error))
        expected = {'code': 400,
                    'error': {'message': 'Error with \n newline',
                              'traceback': None,
                              'type': 'ErrorWithNewline'},
                    'explanation': ('The server could not comply with the '
                                    'request since it is either malformed '
                                    'or otherwise incorrect.'),
                    'title': 'Bad Request'}
        self.assertEqual(expected, msg)

    def test_http_exception_with_traceback(self):
        wrapper = fault.FaultWrapper(None)
        newline_error = ErrorWithNewline(
            'Error with \n newline\nTraceback (most recent call last):\nFoo')
        msg = wrapper._error(heat_exc.HTTPExceptionDisguise(newline_error))
        expected = {'code': 400,
                    'error': {'message': 'Error with \n newline',
                              'traceback': None,
                              'type': 'ErrorWithNewline'},
                    'explanation': ('The server could not comply with the '
                                    'request since it is either malformed '
                                    'or otherwise incorrect.'),
                    'title': 'Bad Request'}
        self.assertEqual(expected, msg)

    def test_openstack_exception_with_kwargs(self):
        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(heat_exc.EntityNotFound(entity='Stack', name='a'))
        expected = {'code': 404,
                    'error': {'message': 'The Stack (a) could not be found.',
                              'traceback': None,
                              'type': 'EntityNotFound'},
                    'explanation': 'The resource could not be found.',
                    'title': 'Not Found'}
        self.assertEqual(expected, msg)

    def test_openstack_exception_without_kwargs(self):
        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(heat_exc.StackResourceLimitExceeded())
        expected = {'code': 500,
                    'error': {'message': 'Maximum resources '
                                         'per stack exceeded.',
                              'traceback': None,
                              'type': 'StackResourceLimitExceeded'},
                    'explanation': 'The server has either erred or is '
                                   'incapable of performing the requested '
                                   'operation.',
                    'title': 'Internal Server Error'}
        self.assertEqual(expected, msg)

    def test_exception_with_non_ascii_chars(self):
        # We set debug to true to test the code path for serializing traces too
        cfg.CONF.set_override('debug', True)
        msg = u'Error with non-ascii chars \x80'

        class TestException(heat_exc.HeatException):
            msg_fmt = msg

        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(TestException())
        expected = {'code': 500,
                    'error': {'message': u'Error with non-ascii chars \x80',
                              'traceback': 'None\n',
                              'type': 'TestException'},
                    'explanation': ('The server has either erred or is '
                                    'incapable of performing the requested '
                                    'operation.'),
                    'title': 'Internal Server Error'}
        self.assertEqual(expected, msg)

    def test_remote_exception(self):
        # We want tracebacks
        cfg.CONF.set_override('debug', True)
        error = heat_exc.EntityNotFound(entity='Stack', name='a')
        exc_info = (type(error), error, None)
        serialized = rpc_common.serialize_remote_exception(exc_info)
        remote_error = rpc_common.deserialize_remote_exception(
            serialized, ["heat.common.exception"])
        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(remote_error)
        expected_message, expected_traceback = str(
            remote_error).split('\n', 1)
        expected = {'code': 404,
                    'error': {'message': expected_message,
                              'traceback': expected_traceback,
                              'type': 'EntityNotFound'},
                    'explanation': 'The resource could not be found.',
                    'title': 'Not Found'}
        self.assertEqual(expected, msg)

    def remote_exception_helper(self, name, error):
        error.args = ()
        exc_info = (type(error), error, None)

        serialized = rpc_common.serialize_remote_exception(exc_info)
        remote_error = rpc_common.deserialize_remote_exception(
            serialized, name)
        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(remote_error)
        expected = {'code': 500,
                    'error': {'traceback': None,
                              'type': 'RemoteError'},
                    'explanation': msg['explanation'],
                    'title': 'Internal Server Error'}
        self.assertEqual(expected, msg)

    def test_all_remote_exceptions(self):
        for name, obj in inspect.getmembers(
                heat_exc, lambda x: inspect.isclass(x) and issubclass(
                    x, heat_exc.HeatException)):

            if '__init__' in obj.__dict__:
                if obj == heat_exc.HeatException:  # manually ignore baseclass
                    continue
                elif obj == heat_exc.Error:
                    error = obj('Error')
                elif obj == heat_exc.NotFound:
                    error = obj()
                elif obj == heat_exc.ResourceFailure:
                    exc = heat_exc.Error(_('Error'))
                    error = obj(exc, None, 'CREATE')
                elif obj == heat_exc.ResourcePropertyConflict:
                    error = obj('%s' % 'a test prop')
                else:
                    continue
                self.remote_exception_helper(name, error)
                continue

            if hasattr(obj, 'msg_fmt'):
                kwargs = {}
                spec_names = re.findall(r'%\((\w+)\)([cdeEfFgGinorsxX])',
                                        obj.msg_fmt)

                for key, convtype in spec_names:
                    if convtype == 'r' or convtype == 's':
                        kwargs[key] = '"' + key + '"'
                    else:
                        # this is highly unlikely
                        raise Exception("test needs additional conversion"
                                        " type added due to %s exception"
                                        " using '%c' specifier" % (obj,
                                                                   convtype))

                error = obj(**kwargs)
                self.remote_exception_helper(name, error)

    def test_should_not_ignore_parent_classes(self):
        wrapper = fault.FaultWrapper(None)

        msg = wrapper._error(StackNotFoundChild(entity='Stack', name='a'))
        expected = {'code': 404,
                    'error': {'message': 'The Stack (a) could not be found.',
                              'traceback': None,
                              'type': 'StackNotFoundChild'},
                    'explanation': 'The resource could not be found.',
                    'title': 'Not Found'}
        self.assertEqual(expected, msg)

    def test_internal_server_error_when_exeption_and_parents_not_mapped(self):
        wrapper = fault.FaultWrapper(None)

        class NotMappedException(Exception):
            pass

        msg = wrapper._error(NotMappedException('A message'))
        expected = {'code': 500,
                    'error': {'traceback': None,
                              'type': 'NotMappedException'},
                    'explanation': ('The server has either erred or is '
                                    'incapable of performing the requested '
                                    'operation.'),
                    'title': 'Internal Server Error'}
        self.assertEqual(expected, msg)

    def test_should_not_ignore_parent_classes_even_for_remote_ones(self):
        # We want tracebacks
        cfg.CONF.set_override('debug', True)

        error = StackNotFoundChild(entity='Stack', name='a')
        exc_info = (type(error), error, None)
        serialized = rpc_common.serialize_remote_exception(exc_info)
        remote_error = rpc_common.deserialize_remote_exception(
            serialized, ["heat.tests.test_fault_middleware"])

        wrapper = fault.FaultWrapper(None)
        msg = wrapper._error(remote_error)
        expected_message, expected_traceback = str(
            remote_error).split('\n', 1)
        expected = {'code': 404,
                    'error': {'message': expected_message,
                              'traceback': expected_traceback,
                              'type': 'StackNotFoundChild'},
                    'explanation': 'The resource could not be found.',
                    'title': 'Not Found'}
        self.assertEqual(expected, msg)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

from heat.common import exception
from heat.common import password_gen
from heat.engine.clients import progress
from heat.engine.resources import stack_user

cfg.CONF.import_opt('max_server_name_length', 'heat.common.config')

LOG = logging.getLogger(__name__)


class BaseServer(stack_user.StackUser):
    """Base Server resource."""

    physical_resource_name_limit = cfg.CONF.max_server_name_length

    entity = 'servers'

    def __init__(self, name, json_snippet, stack):
        super(BaseServer, self).__init__(name, json_snippet, stack)
        self.default_collectors = []

    def _server_name(self):
        name = self.properties[self.NAME]
        if name:
            return name

        return self.physical_resource_name()

    def _container_and_object_name(self, props):
        deployment_swift_data = props.get(
            self.DEPLOYMENT_SWIFT_DATA,
            self.properties[self.DEPLOYMENT_SWIFT_DATA])
        container_name = deployment_swift_data[self.CONTAINER]
        if container_name is None:
            container_name = self.physical_resource_name()

        object_name = deployment_swift_data[self.OBJECT]
        if object_name is None:
            object_name = self.data().get('metadata_object_name')
        if object_name is None:
            object_name = str(uuid.uuid4())

        return container_name, object_name

    def _get_region_name(self):
        return self.client_plugin()._get_region_name()

    def _populate_deployments_metadata(self, meta, props):
        meta['deployments'] = meta.get('deployments', [])
        meta['os-collect-config'] = meta.get('os-collect-config', {})
        occ = meta['os-collect-config']
        collectors = list(self.default_collectors)
        occ['collectors'] = collectors

        # set existing values to None to override any boot-time config
        occ_keys = ('heat', 'zaqar', 'cfn', 'request')
        for occ_key in occ_keys:
            if occ_key not in occ:
                continue
            existing = occ[occ_key]
            for k in existing:
                existing[k] = None

        queue_id = self.data().get('metadata_queue_id')
        if self.transport_poll_server_heat(props):
            occ.update({'heat': {
                'user_id': self._get_user_id(),
                'password': self.password,
                'auth_url': self.keystone().server_keystone_endpoint_url(
                    fallback_endpoint=self.context.auth_url),
                'project_id': self.stack.stack_user_project_id,
                'stack_id': self.stack.identifier().stack_path(),
                'resource_name': self.name,
                'region_name': self._get_region_name()}})
            collectors.append('heat')

        elif self.transport_zaqar_message(props):
            queue_id = queue_id or self.physical_resource_name()
            occ.update({'zaqar': {
                'user_id': self._get_user_id(),
                'password': self.password,
                'auth_url': self.keystone().server_keystone_endpoint_url(
                    fallback_endpoint=self.context.auth_url),
                'project_id': self.stack.stack_user_project_id,
                'queue_id': queue_id,
                'region_name': self._get_region_name()}})
            collectors.append('zaqar')

        elif self.transport_poll_server_cfn(props):
            heat_client_plugin = self.stack.clients.client_plugin('heat')
            config_url = heat_client_plugin.get_cfn_metadata_server_url()
            occ.update({'cfn': {
                'metadata_url': config_url,
                'access_key_id': self.access_key,
                'secret_access_key': self.secret_key,
                'stack_name': self.stack.name,
                'path': '%s.Metadata' % self.name}})
            collectors.append('cfn')

        elif self.transport_poll_temp_url(props):
            container_name, object_name = self._container_and_object_name(
                props)

            self.client('swift').put_container(container_name)

            url = self.client_plugin('swift').get_temp_url(
                container_name, object_name, method='GET')
            put_url = self.client_plugin('swift').get_temp_url(
                container_name, object_name)
            self.data_set('metadata_put_url', put_url)
            self.data_set('metadata_object_name', object_name)

            collectors.append('request')
            occ.update({'request': {'metadata_url': url}})

        collectors.append('local')
        self.metadata_set(meta)

        # push replacement polling config to any existing push-based sources
        if queue_id:
            zaqar_plugin = self.client_plugin('zaqar')
            zaqar = zaqar_plugin.create_for_tenant(
                self.stack.stack_user_project_id, self._user_token())
            queue = zaqar.queue(queue_id)
            queue.post({'body': meta, 'ttl': zaqar_plugin.DEFAULT_TTL})
            self.data_set('metadata_queue_id', queue_id)

        object_name = self.data().get('metadata_object_name')
        if object_name:
            container_name, object_name = self._container_and_object_name(
                props)
            self.client('swift').put_object(
                container_name, object_name, jsonutils.dumps(meta))

    def _create_transport_credentials(self, props):
        if self.transport_poll_server_cfn(props):
            self._create_user()
            self._create_keypair()

        elif (self.transport_poll_server_heat(props) or
              self.transport_zaqar_message(props)):
            if self.password is None:
                self.password = password_gen.generate_openstack_password()
            self._create_user()

        self._register_access_key()

    @property
    def access_key(self):
        return self.data().get('access_key')

    @property
    def secret_key(self):
        return self.data().get('secret_key')

    @property
    def password(self):
        return self.data().get('password')

    @password.setter
    def password(self, password):
        if password is None:
            self.data_delete('password')
        else:
            self.data_set('password', password, True)

    def transport_poll_server_cfn(self, props):
        return props[
            self.SOFTWARE_CONFIG_TRANSPORT] == self.POLL_SERVER_CFN

    def transport_poll_server_heat(self, props):
        return props[
            self.SOFTWARE_CONFIG_TRANSPORT] == self.POLL_SERVER_HEAT

    def transport_poll_temp_url(self, props):
        return props[
            self.SOFTWARE_CONFIG_TRANSPORT] == self.POLL_TEMP_URL

    def transport_zaqar_message(self, props):
        return props[
            self.SOFTWARE_CONFIG_TRANSPORT] == self.ZAQAR_MESSAGE

    def check_create_complete(self, server_id):
        return True

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        if name == self.NAME_ATTR:
            return self._server_name()
        if name == self.OS_COLLECT_CONFIG:
            return self.metadata_get().get('os-collect-config', {})

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if tmpl_diff.metadata_changed():
            # If SOFTWARE_CONFIG user_data_format is enabled we require
            # the "deployments" and "os-collect-config" keys for Deployment
            # polling.  We can attempt to merge the occ data, but any
            # metadata update containing deployments will be discarded.
            new_md = json_snippet.metadata()
            if self.user_data_software_config():
                metadata = self.metadata_get(True) or {}
                new_occ_md = new_md.get('os-collect-config', {})
                occ_md = metadata.get('os-collect-config', {})
                occ_md.update(new_occ_md)
                new_md['os-collect-config'] = occ_md
                deployment_md = metadata.get('deployments', [])
                new_md['deployments'] = deployment_md
            self.metadata_set(new_md)

        updaters = []

        if self.SOFTWARE_CONFIG_TRANSPORT in prop_diff:
            self._update_software_config_transport(prop_diff)

        # NOTE(pas-ha) optimization is possible (starting first task
        # right away), but we'd rather not, as this method already might
        # have called several APIs
        return updaters

    def _update_software_config_transport(self, prop_diff):
        if not self.user_data_software_config():
            return
        self._delete_queue()
        self._delete_temp_url()

        metadata = self.metadata_get(True) or {}
        self._create_transport_credentials(prop_diff)
        self._populate_deployments_metadata(metadata, prop_diff)

    def metadata_update(self, new_metadata=None):
        """Refresh the metadata if new_metadata is None."""
        if new_metadata is None:
            # Re-resolve the template metadata and merge it with the
            # current resource metadata.  This is necessary because the
            # attributes referenced in the template metadata may change
            # and the resource itself adds keys to the metadata which
            # are not specified in the template (e.g the deployments data)
            meta = self.metadata_get(refresh=True) or {}
            tmpl_meta = self.t.metadata()
            meta.update(tmpl_meta)
            self.metadata_set(meta)

    @staticmethod
    def _check_maximum(count, maximum, msg):
        """Check a count against a maximum.

        Unless maximum is -1 which indicates that there is no limit.
        """
        if maximum != -1 and count > maximum:
            raise exception.StackValidationFailed(message=msg)

    def _delete_temp_url(self):
        object_name = self.data().get('metadata_object_name')
        if not object_name:
            return
        endpoint_exists = self.client_plugin().does_endpoint_exist(
            'swift', 'object-store')
        if endpoint_exists:
            with self.client_plugin('swift').ignore_not_found:
                container = self.properties[self.DEPLOYMENT_SWIFT_DATA].get(
                    'container')
                container = container or self.physical_resource_name()
                swift = self.client('swift')
                swift.delete_object(container, object_name)
                headers = swift.head_container(container)
                if int(headers['x-container-object-count']) == 0:
                    swift.delete_container(container)
        self.data_delete('metadata_object_name')
        self.data_delete('metadata_put_url')

    def _delete_queue(self):
        queue_id = self.data().get('metadata_queue_id')
        if not queue_id:
            return
        endpoint_exists = self.client_plugin().does_endpoint_exist(
            'zaqar', 'messaging')
        if endpoint_exists:
            client_plugin = self.client_plugin('zaqar')
            zaqar = client_plugin.create_for_tenant(
                self.stack.stack_user_project_id, self._user_token())
            with client_plugin.ignore_not_found:
                zaqar.queue(queue_id).delete()
        self.data_delete('metadata_queue_id')

    def handle_snapshot_delete(self, state):

        if state[1] != self.FAILED and self.resource_id:
            image_id = self.client().servers.create_image(
                self.resource_id, self.physical_resource_name())
            return progress.ServerDeleteProgress(
                self.resource_id, image_id, False)
        return self._delete()

    def handle_delete(self):

        return self._delete()

    def check_delete_complete(self, prg):
        if not prg:
            return True

    def _show_resource(self):
        rsrc_dict = super(BaseServer, self)._show_resource()
        rsrc_dict.setdefault(
            self.OS_COLLECT_CONFIG,
            self.metadata_get().get('os-collect-config', {}))
        return rsrc_dict



#
# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path

import ddt

from oslo_config import fixture as config_fixture
from oslo_policy import policy as base_policy

from heat.common import exception
from heat.common import policy
from heat.tests import common
from heat.tests import utils

policy_path = os.path.dirname(os.path.realpath(__file__)) + "/policy/"


@ddt.ddt
class TestPolicyEnforcer(common.HeatTestCase):

    def setUp(self):
        super(TestPolicyEnforcer, self).setUp(mock_resource_policy=False)
        self.fixture = self.useFixture(config_fixture.Config())
        self.fixture.conf(args=['--config-dir', policy_path])

    def get_policy_file(self, filename):
        return policy_path + filename

    def _get_context(self, persona):
        if persona == "system_admin":
            ctx = utils.dummy_system_admin_context()
        elif persona == "system_reader":
            ctx = utils.dummy_system_reader_context()
        elif persona == "project_admin":
            ctx = utils.dummy_context(roles=['admin', 'member', 'reader'])
        elif persona == "project_member":
            ctx = utils.dummy_context(roles=['member', 'reader'])
        elif persona == "project_reader":
            ctx = utils.dummy_context(roles=['reader'])
        elif persona == "stack_user":
            ctx = utils.dummy_context(roles=['heat_stack_user'])
        elif persona == "anyone":
            ctx = utils.dummy_context(roles=['foobar'])
        else:
            self.fail("Persona [{}] not found".format(persona))
        return ctx

    def _test_legacy_rbac_policies(self, **kwargs):
        scope = kwargs.get("scope")
        actions = kwargs.get("actions")
        allowed_personas = kwargs.get("allowed", [])
        denied_personas = kwargs.get("denied", [])
        self._test_policy_allowed(scope, actions, allowed_personas)
        self._test_policy_notallowed(scope, actions, denied_personas)

    @ddt.file_data('policy/test_acl_personas.yaml')
    @ddt.unpack
    def test_legacy_rbac_policies(self, **kwargs):
        self._test_legacy_rbac_policies(**kwargs)

    @ddt.file_data('policy/test_deprecated_access.yaml')
    @ddt.unpack
    def test_deprecated_policies(self, **kwargs):
        self._test_legacy_rbac_policies(**kwargs)

    @ddt.file_data('policy/test_acl_personas.yaml')
    @ddt.unpack
    def test_secure_rbac_policies(self, **kwargs):
        self.fixture.config(group='oslo_policy', enforce_scope=True)
        self.fixture.config(group='oslo_policy', enforce_new_defaults=True)
        scope = kwargs.get("scope")
        actions = kwargs.get("actions")
        allowed_personas = kwargs.get("allowed", [])
        denied_personas = kwargs.get("denied", [])
        self._test_policy_allowed(scope, actions, allowed_personas)
        self._test_policy_notallowed(scope, actions, denied_personas)

    def _test_policy_allowed(self, scope, actions, personas):
        enforcer = policy.Enforcer(scope=scope)
        for persona in personas:
            ctx = self._get_context(persona)
            for action in actions:
                # Everything should be allowed
                enforcer.enforce(
                    ctx,
                    action,
                    target={"project_id": "test_tenant_id"},
                    is_registered_policy=True
                )

    def _test_policy_notallowed(self, scope, actions, personas):
        enforcer = policy.Enforcer(scope=scope)
        for persona in personas:
            ctx = self._get_context(persona)
            for action in actions:
                # Everything should raise the default exception.Forbidden
                self.assertRaises(
                    exception.Forbidden,
                    enforcer.enforce, ctx,
                    action,
                    target={"project_id": "test_tenant_id"},
                    is_registered_policy=True)

    def test_set_rules_overwrite_true(self):
        enforcer = policy.Enforcer()
        enforcer.load_rules(True)
        enforcer.set_rules({'test_heat_rule': 1}, True)
        self.assertEqual({'test_heat_rule': 1}, enforcer.enforcer.rules)

    def test_set_rules_overwrite_false(self):
        enforcer = policy.Enforcer()
        enforcer.load_rules(True)
        enforcer.load_rules(True)
        enforcer.set_rules({'test_heat_rule': 1}, False)
        self.assertIn('test_heat_rule', enforcer.enforcer.rules)

    def test_load_rules_force_reload_true(self):
        enforcer = policy.Enforcer()
        enforcer.load_rules(True)
        enforcer.set_rules({'test_heat_rule': 'test'})
        enforcer.load_rules(True)
        self.assertNotIn({'test_heat_rule': 'test'}, enforcer.enforcer.rules)

    def test_load_rules_force_reload_false(self):
        enforcer = policy.Enforcer()
        enforcer.load_rules(True)
        enforcer.load_rules(True)
        enforcer.set_rules({'test_heat_rule': 'test'})
        enforcer.load_rules(False)
        self.assertIn('test_heat_rule', enforcer.enforcer.rules)

    def test_no_such_action(self):
        ctx = utils.dummy_context(roles=['not_a_stack_user'])
        enforcer = policy.Enforcer(scope='cloudformation')
        action = 'no_such_action'
        msg = 'cloudformation:no_such_action has not been registered'
        self.assertRaisesRegex(base_policy.PolicyNotRegistered,
                               msg,
                               enforcer.enforce,
                               ctx, action,
                               None, None,
                               True)

    def test_check_admin(self):
        enforcer = policy.Enforcer()

        ctx = utils.dummy_context(roles=[])
        self.assertFalse(enforcer.check_is_admin(ctx))

        ctx = utils.dummy_context(roles=['not_admin'])
        self.assertFalse(enforcer.check_is_admin(ctx))

        ctx = utils.dummy_context(roles=['admin'])
        self.assertTrue(enforcer.check_is_admin(ctx))

    def test_enforce_creds(self):
        enforcer = policy.Enforcer()
        ctx = utils.dummy_context(roles=['admin'])
        self.assertTrue(enforcer.check_is_admin(ctx))

    def test_resource_default_rule(self):
        context = utils.dummy_context(roles=['non-admin'])
        enforcer = policy.ResourceEnforcer()
        res_type = "OS::Test::NotInPolicy"
        self.assertTrue(enforcer.enforce(context, res_type,
                                         is_registered_policy=True))

    def test_resource_enforce_success(self):
        context = utils.dummy_context(roles=['admin'])
        enforcer = policy.ResourceEnforcer()
        res_type = "OS::Keystone::User"
        self.assertTrue(enforcer.enforce(context, res_type,
                                         is_registered_policy=True))

    def test_resource_enforce_fail(self):
        context = utils.dummy_context(roles=['non-admin'])
        enforcer = policy.ResourceEnforcer()
        res_type = "OS::Nova::Quota"
        ex = self.assertRaises(exception.Forbidden,
                               enforcer.enforce,
                               context, res_type,
                               None, None,
                               True)
        self.assertIn(res_type, ex.message)

    def test_resource_wildcard_enforce_fail(self):
        context = utils.dummy_context(roles=['non-admin'])
        enforcer = policy.ResourceEnforcer()
        res_type = "OS::Keystone::User"
        ex = self.assertRaises(exception.Forbidden,
                               enforcer.enforce,
                               context, res_type,
                               None, None,
                               True)

        self.assertIn(res_type.split("::", 1)[0], ex.message)

    def test_resource_enforce_returns_false(self):
        context = utils.dummy_context(roles=['non-admin'])
        enforcer = policy.ResourceEnforcer(exc=None)
        res_type = "OS::Keystone::User"
        self.assertFalse(enforcer.enforce(context, res_type,
                                          is_registered_policy=True))
        self.assertIsNotNone(enforcer.enforce(context, res_type,
                                              is_registered_policy=True))

    def test_resource_enforce_exc_on_false(self):
        context = utils.dummy_context(roles=['non-admin'])
        enforcer = policy.ResourceEnforcer()
        res_type = "OS::Keystone::User"
        ex = self.assertRaises(exception.Forbidden,
                               enforcer.enforce,
                               context, res_type,
                               None, None,
                               True)

        self.assertIn(res_type, ex.message)

    def test_resource_enforce_override_deny_admin(self):
        context = utils.dummy_context(roles=['admin'])
        enforcer = policy.ResourceEnforcer(
            policy_file=self.get_policy_file('resources.json'))
        res_type = "OS::Cinder::Quota"
        ex = self.assertRaises(exception.Forbidden,
                               enforcer.enforce,
                               context, res_type,
                               None, None,
                               True)
        self.assertIn(res_type, ex.message)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.engine.resources.openstack.designate import zone
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils


sample_template = {
    'heat_template_version': '2015-04-30',
    'resources': {
        'test_resource': {
            'type': 'OS::Designate::Zone',
            'properties': {
                'name': 'test-zone.com',
                'description': 'Test zone',
                'ttl': 3600,
                'email': 'abc@test-zone.com',
                'type': 'PRIMARY',
                'masters': []
            }
        }
    }
}


class DesignateZoneTest(common.HeatTestCase):

    def setUp(self):
        super(DesignateZoneTest, self).setUp()

        self.ctx = utils.dummy_context()

        self.stack = stack.Stack(
            self.ctx, 'test_stack',
            template.Template(sample_template)
        )

        self.test_resource = self.stack['test_resource']

        # Mock client plugin
        self.test_client_plugin = mock.MagicMock()
        self.test_resource.client_plugin = mock.MagicMock(
            return_value=self.test_client_plugin)

        # Mock client
        self.test_client = mock.MagicMock()
        self.test_resource.client = mock.MagicMock(
            return_value=self.test_client)

    def _get_mock_resource(self):
        value = {}
        value['id'] = '477e8273-60a7-4c41-b683-fdb0bc7cd152'
        value['serial'] = '1434596972'

        return value

    def test_resource_handle_create(self):
        mock_zone_create = self.test_client.zones.create
        mock_resource = self._get_mock_resource()
        mock_zone_create.return_value = mock_resource

        # validate the properties
        self.assertEqual(
            'test-zone.com',
            self.test_resource.properties.get(zone.DesignateZone.NAME))
        self.assertEqual(
            'Test zone',
            self.test_resource.properties.get(
                zone.DesignateZone.DESCRIPTION))
        self.assertEqual(
            3600,
            self.test_resource.properties.get(zone.DesignateZone.TTL))
        self.assertEqual(
            'abc@test-zone.com',
            self.test_resource.properties.get(zone.DesignateZone.EMAIL))

        self.assertEqual(
            'PRIMARY',
            self.test_resource.properties.get(zone.DesignateZone.TYPE))

        self.assertEqual(
            [],
            self.test_resource.properties.get(zone.DesignateZone.MASTERS))

        self.test_resource.data_set = mock.Mock()
        self.test_resource.handle_create()

        args = dict(
            name='test-zone.com',
            description='Test zone',
            ttl=3600,
            email='abc@test-zone.com',
            type_='PRIMARY'
        )

        mock_zone_create.assert_called_once_with(**args)
        # validate physical resource id
        self.assertEqual(mock_resource['id'], self.test_resource.resource_id)

    def _mock_check_status_active(self):
        self.test_client.zones.get.side_effect = [
            {'status': 'PENDING'},
            {'status': 'ACTIVE'},
            {'status': 'ERROR'}
        ]

    def test_check_create_complete(self):
        self._mock_check_status_active()
        self.assertFalse(self.test_resource.check_create_complete())
        self.assertTrue(self.test_resource.check_create_complete())
        ex = self.assertRaises(exception.ResourceInError,
                               self.test_resource.check_create_complete)
        self.assertIn('Error in zone',
                      ex.message)

    def _test_resource_validate(self, type_, prp):
        def _side_effect(key):
            if key == prp:
                return None
            if key == zone.DesignateZone.TYPE:
                return type_
            else:
                return sample_template['resources'][
                    'test_resource']['properties'][key]

        self.test_resource.properties = mock.MagicMock()
        self.test_resource.properties.get.side_effect = _side_effect
        self.test_resource.properties.__getitem__.side_effect = _side_effect

        ex = self.assertRaises(exception.StackValidationFailed,
                               self.test_resource.validate)
        self.assertEqual('Property %s is required for zone type %s' %
                         (prp, type_),
                         ex.message)

    def test_resource_validate_primary(self):
        self._test_resource_validate(zone.DesignateZone.PRIMARY,
                                     zone.DesignateZone.EMAIL)

    def test_resource_validate_secondary(self):
        self._test_resource_validate(zone.DesignateZone.SECONDARY,
                                     zone.DesignateZone.MASTERS)

    def test_resource_handle_update(self):
        mock_zone_update = self.test_client.zones.update
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        prop_diff = {zone.DesignateZone.EMAIL: 'xyz@test-zone.com',
                     zone.DesignateZone.DESCRIPTION: 'updated description',
                     zone.DesignateZone.TTL: 4200}

        self.test_resource.handle_update(json_snippet=None,
                                         tmpl_diff=None,
                                         prop_diff=prop_diff)

        args = dict(
            description='updated description',
            ttl=4200,
            email='xyz@test-zone.com'
        )
        mock_zone_update.assert_called_once_with(
            self.test_resource.resource_id,
            args)

    def test_check_update_complete(self):
        self._mock_check_status_active()
        self.assertFalse(self.test_resource.check_update_complete())
        self.assertTrue(self.test_resource.check_update_complete())
        ex = self.assertRaises(exception.ResourceInError,
                               self.test_resource.check_update_complete)
        self.assertIn('Error in zone',
                      ex.message)

    def test_check_delete_complete(self):
        self._mock_check_status_active()
        self.assertFalse(self.test_resource.check_delete_complete(
            self._get_mock_resource()['id']
        ))
        self.assertTrue(self.test_resource.check_delete_complete(
            self._get_mock_resource()['id']
        ))
        ex = self.assertRaises(exception.ResourceInError,
                               self.test_resource.check_delete_complete,
                               self._get_mock_resource()['id'])
        self.assertIn('Error in zone',
                      ex.message)

    def test_resolve_attributes(self):
        mock_zone = self._get_mock_resource()
        self.test_resource.resource_id = mock_zone['id']
        self.test_client.zones.get.return_value = mock_zone
        self.assertEqual(
            mock_zone['serial'],
            self.test_resource._resolve_attribute(zone.DesignateZone.SERIAL))
        self.test_client.zones.get.assert_called_once_with(
            self.test_resource.resource_id
        )


class DesignateSecondaryZoneTest(common.HeatTestCase):

    def setUp(self):
        super(DesignateSecondaryZoneTest, self).setUp()

        self.ctx = utils.dummy_context()

        self.primaries = ['::1']

    def _get_mock_resource(self):
        value = {}
        value['id'] = '477e8273-60a7-4c41-b683-fdb0bc7cd152'
        value['serial'] = '1434596972'

        return value

    def test_masters(self):
        self.do_test({
            'heat_template_version': '2015-04-30',
            'resources': {
                'test_resource': {
                    'type': 'OS::Designate::Zone',
                    'properties': {
                        'name': 'test-zone.com',
                        'description': 'Test zone',
                        'ttl': 3600,
                        'email': 'abc@test-zone.com',
                        'type': 'SECONDARY',
                        'masters': self.primaries,
                    }
                }
            }
        })

    def test_primaries(self):
        self.do_test({
            'heat_template_version': '2015-04-30',
            'resources': {
                'test_resource': {
                    'type': 'OS::Designate::Zone',
                    'properties': {
                        'name': 'test-zone.com',
                        'description': 'Test zone',
                        'ttl': 3600,
                        'email': 'abc@test-zone.com',
                        'type': 'SECONDARY',
                        'primaries': self.primaries,
                    }
                }
            }
        })

    def do_test(self, sample_template):
        self.stack = stack.Stack(
            self.ctx, 'test_stack',
            template.Template(sample_template)
        )

        self.test_resource = self.stack['test_resource']

        # Mock client plugin
        self.test_client_plugin = mock.MagicMock()
        self.test_resource.client_plugin = mock.MagicMock(
            return_value=self.test_client_plugin)

        # Mock client
        self.test_client = mock.MagicMock()
        self.test_resource.client = mock.MagicMock(
            return_value=self.test_client)

        mock_zone_create = self.test_client.zones.create
        mock_resource = self._get_mock_resource()
        mock_zone_create.return_value = mock_resource

        self.test_resource.data_set = mock.Mock()
        self.test_resource.handle_create()

        args = dict(
            name='test-zone.com',
            description='Test zone',
            ttl=3600,
            email='abc@test-zone.com',
            type_='SECONDARY',
            masters=self.primaries
        )

        mock_zone_create.assert_called_once_with(**args)
        # validate physical resource id
        self.assertEqual(mock_resource['id'], self.test_resource.resource_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from heat_integrationtests.functional import functional_base


class SimpleStackValidationTest(functional_base.FunctionalTestsBase):

    def test_validate_json_content(self):
        template = u'''
heat_template_version: rocky
resources:
  server:
    type: OS::Heat::TestResource
    properties:
      value: =%da
'''
        self.stack_create(template=template,
                          expected_status='CREATE_COMPLETE')


class StackValidationTest(functional_base.FunctionalTestsBase):

    def setUp(self):
        super(StackValidationTest, self).setUp()
        if not self.conf.minimal_image_ref:
            raise self.skipException("No image configured to test")

        if not self.conf.minimal_instance_type:
            raise self.skipException(
                "No minimal_instance_type configured to test")

        self.assign_keypair()

    def test_stack_validate_provider_references_parent_resource(self):
        template = '''
heat_template_version: 2014-10-16
parameters:
  keyname:
    type: string
  flavor:
    type: string
  image:
    type: string
  network:
    type: string
resources:
  config:
    type: My::Config
    properties:
        server: {get_resource: server}

  server:
    type: OS::Nova::Server
    properties:
      image: {get_param: image}
      flavor: {get_param: flavor}
      key_name: {get_param: keyname}
      networks: [{network: {get_param: network} }]
      user_data_format: SOFTWARE_CONFIG

'''
        config_template = '''
heat_template_version: 2014-10-16
parameters:
  server:
    type: string
resources:
  config:
    type: OS::Heat::SoftwareConfig

  deployment:
    type: OS::Heat::SoftwareDeployment
    properties:
      config:
        get_resource: config
      server:
        get_param: server
'''
        files = {'provider.yaml': config_template}
        env = {'resource_registry':
               {'My::Config': 'provider.yaml'}}
        parameters = {'keyname': self.keypair_name,
                      'flavor': self.conf.minimal_instance_type,
                      'image': self.conf.minimal_image_ref,
                      'network': self.conf.fixed_network_name}
        # Note we don't wait for CREATE_COMPLETE, because we're using a
        # minimal image without the tools to apply the config.
        # The point of the test is just to prove that validation won't
        # falsely prevent stack creation starting, ref bug #1407100
        # Note that we can be sure My::Config will stay IN_PROGRESS as
        # there's no signal sent to the deployment
        self.stack_create(template=template,
                          files=files,
                          environment=env,
                          parameters=parameters,
                          expected_status='CREATE_IN_PROGRESS')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat_integrationtests.functional import functional_base


TEMPLATE = '''
    heat_template_version: 2015-04-30
    parameters:
      p0:
        type: string
        default: CORRECT
      p1:
        type: string
        default: INCORRECT
      p2:
        type: string
        default: INCORRECT
    resources:
      r1:
        type: test::R1
      r2:
        type: test::R2
      r3a:
        type: test::R3
      r3b:
        type: test::R3
'''

ENV_1 = '''
    parameters:
      p1: CORRECT
      p2: INCORRECT-E1
    resource_registry:
      test::R1: OS::Heat::RandomString
      test::R2: BROKEN
      test::R3: OS::Heat::None
'''

ENV_2 = '''
    parameters:
      p2: CORRECT
    resource_registry:
      test::R2: OS::Heat::RandomString
      resources:
        r3b:
          test::R3: OS::Heat::RandomString
'''


class EnvironmentMergingTests(functional_base.FunctionalTestsBase):

    def test_server_environment_merging(self):

        # Setup
        files = {'env1.yaml': ENV_1, 'env2.yaml': ENV_2}
        environment_files = ['env1.yaml', 'env2.yaml']

        # Test
        stack_id = self.stack_create(stack_name='env_merge',
                                     template=TEMPLATE,
                                     files=files,
                                     environment_files=environment_files)

        # Verify

        # Since there is no environment show, the registry overriding
        # is partially verified by there being no error. If it wasn't
        # working, test::R2 would remain mapped to BROKEN in env1.

        # Sanity check
        resources = self.list_resources(stack_id)
        self.assertEqual(4, len(resources))

        # Verify the parameters are correctly set
        stack = self.client.stacks.get(stack_id)
        self.assertEqual('CORRECT', stack.parameters['p0'])
        self.assertEqual('CORRECT', stack.parameters['p1'])
        self.assertEqual('CORRECT', stack.parameters['p2'])

        # Verify that r3b has been overridden into a RandomString
        # by checking to see that it has a value
        r3b = self.client.resources.get(stack_id, 'r3b')
        r3b_attrs = r3b.attributes
        self.assertIn('value', r3b_attrs)



#
# Copyright (c) 2013 Docker, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import distutils

from oslo_log import log as logging

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support

LOG = logging.getLogger(__name__)

DOCKER_INSTALLED = False
MIN_API_VERSION_MAP = {'read_only': '1.17', 'cpu_shares': '1.8',
                       'devices': '1.14', 'cpu_set': '1.12'}
DEVICE_PATH_REGEX = r"^/dev/[/_\-a-zA-Z0-9]+$"
# conditionally import so tests can work without having the dependency
# satisfied
try:
    import docker
    DOCKER_INSTALLED = True
except ImportError:
    docker = None


class DockerContainer(resource.Resource):

    support_status = support.SupportStatus(
        status=support.UNSUPPORTED,
        message=_('This resource is not supported, use at your own risk.'))

    PROPERTIES = (
        DOCKER_ENDPOINT, HOSTNAME, USER, MEMORY, PORT_SPECS,
        PRIVILEGED, TTY, OPEN_STDIN, STDIN_ONCE, ENV, CMD, DNS,
        IMAGE, VOLUMES, VOLUMES_FROM, PORT_BINDINGS, LINKS, NAME,
        RESTART_POLICY, CAP_ADD, CAP_DROP, READ_ONLY, CPU_SHARES,
        DEVICES, CPU_SET
    ) = (
        'docker_endpoint', 'hostname', 'user', 'memory', 'port_specs',
        'privileged', 'tty', 'open_stdin', 'stdin_once', 'env', 'cmd', 'dns',
        'image', 'volumes', 'volumes_from', 'port_bindings', 'links', 'name',
        'restart_policy', 'cap_add', 'cap_drop', 'read_only', 'cpu_shares',
        'devices', 'cpu_set'
    )

    ATTRIBUTES = (
        INFO, NETWORK_INFO, NETWORK_IP, NETWORK_GATEWAY,
        NETWORK_TCP_PORTS, NETWORK_UDP_PORTS, LOGS, LOGS_HEAD,
        LOGS_TAIL,
    ) = (
        'info', 'network_info', 'network_ip', 'network_gateway',
        'network_tcp_ports', 'network_udp_ports', 'logs', 'logs_head',
        'logs_tail',
    )

    _RESTART_POLICY_KEYS = (
        POLICY_NAME, POLICY_MAXIMUM_RETRY_COUNT,
    ) = (
        'Name', 'MaximumRetryCount',
    )

    _DEVICES_KEYS = (
        PATH_ON_HOST, PATH_IN_CONTAINER, PERMISSIONS
    ) = (
        'path_on_host', 'path_in_container', 'permissions'
    )

    _CAPABILITIES = ['SETPCAP', 'SYS_MODULE', 'SYS_RAWIO', 'SYS_PACCT',
                     'SYS_ADMIN', 'SYS_NICE', 'SYS_RESOURCE', 'SYS_TIME',
                     'SYS_TTY_CONFIG', 'MKNOD', 'AUDIT_WRITE',
                     'AUDIT_CONTROL', 'MAC_OVERRIDE', 'MAC_ADMIN',
                     'NET_ADMIN', 'SYSLOG', 'CHOWN', 'NET_RAW',
                     'DAC_OVERRIDE', 'FOWNER', 'DAC_READ_SEARCH', 'FSETID',
                     'KILL', 'SETGID', 'SETUID', 'LINUX_IMMUTABLE',
                     'NET_BIND_SERVICE', 'NET_BROADCAST', 'IPC_LOCK',
                     'IPC_OWNER', 'SYS_CHROOT', 'SYS_PTRACE', 'SYS_BOOT',
                     'LEASE', 'SETFCAP', 'WAKE_ALARM', 'BLOCK_SUSPEND', 'ALL']

    properties_schema = {
        DOCKER_ENDPOINT: properties.Schema(
            properties.Schema.STRING,
            _('Docker daemon endpoint (by default the local docker daemon '
              'will be used).'),
            default=None
        ),
        HOSTNAME: properties.Schema(
            properties.Schema.STRING,
            _('Hostname of the container.'),
            default=''
        ),
        USER: properties.Schema(
            properties.Schema.STRING,
            _('Username or UID.'),
            default=''
        ),
        MEMORY: properties.Schema(
            properties.Schema.INTEGER,
            _('Memory limit (Bytes).')
        ),
        PORT_SPECS: properties.Schema(
            properties.Schema.LIST,
            _('TCP/UDP ports mapping.'),
            default=None
        ),
        PORT_BINDINGS: properties.Schema(
            properties.Schema.MAP,
            _('TCP/UDP ports bindings.'),
        ),
        LINKS: properties.Schema(
            properties.Schema.MAP,
            _('Links to other containers.'),
        ),
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of the container.'),
        ),
        PRIVILEGED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Enable extended privileges.'),
            default=False
        ),
        TTY: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Allocate a pseudo-tty.'),
            default=False
        ),
        OPEN_STDIN: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Open stdin.'),
            default=False
        ),
        STDIN_ONCE: properties.Schema(
            properties.Schema.BOOLEAN,
            _('If true, close stdin after the 1 attached client disconnects.'),
            default=False
        ),
        ENV: properties.Schema(
            properties.Schema.LIST,
            _('Set environment variables.'),
        ),
        CMD: properties.Schema(
            properties.Schema.LIST,
            _('Command to run after spawning the container.'),
            default=[]
        ),
        DNS: properties.Schema(
            properties.Schema.LIST,
            _('Set custom dns servers.'),
        ),
        IMAGE: properties.Schema(
            properties.Schema.STRING,
            _('Image name.')
        ),
        VOLUMES: properties.Schema(
            properties.Schema.MAP,
            _('Create a bind mount.'),
            default={}
        ),
        VOLUMES_FROM: properties.Schema(
            properties.Schema.LIST,
            _('Mount all specified volumes.'),
            default=''
        ),
        RESTART_POLICY: properties.Schema(
            properties.Schema.MAP,
            _('Restart policies (only supported for API version >= 1.2.0).'),
            schema={
                POLICY_NAME: properties.Schema(
                    properties.Schema.STRING,
                    _('The behavior to apply when the container exits.'),
                    default='no',
                    constraints=[
                        constraints.AllowedValues(['no', 'on-failure',
                                                   'always']),
                    ]
                ),
                POLICY_MAXIMUM_RETRY_COUNT: properties.Schema(
                    properties.Schema.INTEGER,
                    _('A maximum restart count for the '
                      'on-failure policy.'),
                    default=0
                )
            },
            default={},
            support_status=support.SupportStatus(version='2015.1')
        ),
        CAP_ADD: properties.Schema(
            properties.Schema.LIST,
            _('Be used to add kernel capabilities (only supported for '
              'API version >= 1.2.0).'),
            schema=properties.Schema(
                properties.Schema.STRING,
                _('The security features provided by Linux kernels.'),
                constraints=[
                    constraints.AllowedValues(_CAPABILITIES),
                ]
            ),
            default=[],
            support_status=support.SupportStatus(version='2015.1')
        ),
        CAP_DROP: properties.Schema(
            properties.Schema.LIST,
            _('Be used to drop kernel capabilities (only supported for '
              'API version >= 1.2.0).'),
            schema=properties.Schema(
                properties.Schema.STRING,
                _('The security features provided by Linux kernels.'),
                constraints=[
                    constraints.AllowedValues(_CAPABILITIES),
                ]
            ),
            default=[],
            support_status=support.SupportStatus(version='2015.1')
        ),
        READ_ONLY: properties.Schema(
            properties.Schema.BOOLEAN,
            _('If true, mount the container\'s root filesystem '
              'as read only (only supported for API version >= %s).') %
            MIN_API_VERSION_MAP['read_only'],
            default=False,
            support_status=support.SupportStatus(version='2015.1'),
        ),
        CPU_SHARES: properties.Schema(
            properties.Schema.INTEGER,
            _('Relative weight which determines the allocation of the CPU '
              'processing power(only supported for API version >= %s).') %
            MIN_API_VERSION_MAP['cpu_shares'],
            default=0,
            support_status=support.SupportStatus(version='5.0.0'),
        ),
        DEVICES: properties.Schema(
            properties.Schema.LIST,
            _('Device mappings (only supported for API version >= %s).') %
            MIN_API_VERSION_MAP['devices'],
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    PATH_ON_HOST: properties.Schema(
                        properties.Schema.STRING,
                        _('The device path on the host.'),
                        constraints=[
                            constraints.Length(max=255),
                            constraints.AllowedPattern(DEVICE_PATH_REGEX),
                        ],
                        required=True
                    ),
                    PATH_IN_CONTAINER: properties.Schema(
                        properties.Schema.STRING,
                        _('The device path of the container'
                          ' mappings to the host.'),
                        constraints=[
                            constraints.Length(max=255),
                            constraints.AllowedPattern(DEVICE_PATH_REGEX),
                        ],
                    ),
                    PERMISSIONS: properties.Schema(
                        properties.Schema.STRING,
                        _('The permissions of the container to'
                          ' read/write/create the devices.'),
                        constraints=[
                            constraints.AllowedValues(['r', 'w', 'm',
                                                       'rw', 'rm', 'wm',
                                                       'rwm']),
                        ],
                        default='rwm'
                    )
                }
            ),
            default=[],
            support_status=support.SupportStatus(version='5.0.0'),
        ),
        CPU_SET: properties.Schema(
            properties.Schema.STRING,
            _('The CPUs in which to allow execution '
              '(only supported for API version >= %s).') %
            MIN_API_VERSION_MAP['cpu_set'],
            support_status=support.SupportStatus(version='5.0.0'),
        )
    }

    attributes_schema = {
        INFO: attributes.Schema(
            _('Container info.')
        ),
        NETWORK_INFO: attributes.Schema(
            _('Container network info.')
        ),
        NETWORK_IP: attributes.Schema(
            _('Container ip address.')
        ),
        NETWORK_GATEWAY: attributes.Schema(
            _('Container ip gateway.')
        ),
        NETWORK_TCP_PORTS: attributes.Schema(
            _('Container TCP ports.')
        ),
        NETWORK_UDP_PORTS: attributes.Schema(
            _('Container UDP ports.')
        ),
        LOGS: attributes.Schema(
            _('Container logs.')
        ),
        LOGS_HEAD: attributes.Schema(
            _('Container first logs line.')
        ),
        LOGS_TAIL: attributes.Schema(
            _('Container last logs line.')
        ),
    }

    def get_client(self):
        client = None
        if DOCKER_INSTALLED:
            endpoint = self.properties.get(self.DOCKER_ENDPOINT)
            if endpoint:
                client = docker.APIClient(endpoint)
            else:
                client = docker.APIClient()
        return client

    def _parse_networkinfo_ports(self, networkinfo):
        tcp = []
        udp = []
        for port, info in networkinfo['Ports'].items():
            p = port.split('/')
            if not info or len(p) != 2 or 'HostPort' not in info[0]:
                continue
            port = info[0]['HostPort']
            if p[1] == 'tcp':
                tcp.append(port)
            elif p[1] == 'udp':
                udp.append(port)
        return (','.join(tcp), ','.join(udp))

    def _container_networkinfo(self, client, resource_id):
        info = client.inspect_container(self.resource_id)
        networkinfo = info['NetworkSettings']
        ports = self._parse_networkinfo_ports(networkinfo)
        networkinfo['TcpPorts'] = ports[0]
        networkinfo['UdpPorts'] = ports[1]
        return networkinfo

    def _resolve_attribute(self, name):
        if not self.resource_id:
            return
        if name == 'info':
            client = self.get_client()
            return client.inspect_container(self.resource_id)
        if name == 'network_info':
            client = self.get_client()
            networkinfo = self._container_networkinfo(client, self.resource_id)
            return networkinfo
        if name == 'network_ip':
            client = self.get_client()
            networkinfo = self._container_networkinfo(client, self.resource_id)
            return networkinfo['IPAddress']
        if name == 'network_gateway':
            client = self.get_client()
            networkinfo = self._container_networkinfo(client, self.resource_id)
            return networkinfo['Gateway']
        if name == 'network_tcp_ports':
            client = self.get_client()
            networkinfo = self._container_networkinfo(client, self.resource_id)
            return networkinfo['TcpPorts']
        if name == 'network_udp_ports':
            client = self.get_client()
            networkinfo = self._container_networkinfo(client, self.resource_id)
            return networkinfo['UdpPorts']
        if name == 'logs':
            client = self.get_client()
            logs = client.logs(self.resource_id)
            return logs
        if name == 'logs_head':
            client = self.get_client()
            logs = client.logs(self.resource_id)
            return logs.split('\n')[0]
        if name == 'logs_tail':
            client = self.get_client()
            logs = client.logs(self.resource_id)
            return logs.split('\n').pop()

    def handle_create(self):
        create_args = {
            'image': self.properties[self.IMAGE],
            'command': self.properties[self.CMD],
            'hostname': self.properties[self.HOSTNAME],
            'user': self.properties[self.USER],
            'stdin_open': self.properties[self.OPEN_STDIN],
            'tty': self.properties[self.TTY],
            'mem_limit': self.properties[self.MEMORY],
            'ports': self.properties[self.PORT_SPECS],
            'environment': self.properties[self.ENV],
            'dns': self.properties[self.DNS],
            'volumes': self.properties[self.VOLUMES],
            'name': self.properties[self.NAME],
            'cpu_shares': self.properties[self.CPU_SHARES],
            'cpuset': self.properties[self.CPU_SET]
        }
        client = self.get_client()
        client.pull(self.properties[self.IMAGE])
        result = client.create_container(**create_args)
        container_id = result['Id']
        self.resource_id_set(container_id)

        start_args = {}

        if self.properties[self.PRIVILEGED]:
            start_args[self.PRIVILEGED] = True
        if self.properties[self.VOLUMES]:
            start_args['binds'] = self.properties[self.VOLUMES]
        if self.properties[self.VOLUMES_FROM]:
            start_args['volumes_from'] = self.properties[self.VOLUMES_FROM]
        if self.properties[self.PORT_BINDINGS]:
            start_args['port_bindings'] = self.properties[self.PORT_BINDINGS]
        if self.properties[self.LINKS]:
            start_args['links'] = self.properties[self.LINKS]
        if self.properties[self.RESTART_POLICY]:
            start_args['restart_policy'] = self.properties[self.RESTART_POLICY]
        if self.properties[self.CAP_ADD]:
            start_args['cap_add'] = self.properties[self.CAP_ADD]
        if self.properties[self.CAP_DROP]:
            start_args['cap_drop'] = self.properties[self.CAP_DROP]
        if self.properties[self.READ_ONLY]:
            start_args[self.READ_ONLY] = True
        if (self.properties[self.DEVICES] and
                not self.properties[self.PRIVILEGED]):
            start_args['devices'] = self._get_mapping_devices(
                self.properties[self.DEVICES])

        client.start(container_id, **start_args)
        return container_id

    def _get_mapping_devices(self, devices):
        actual_devices = []
        for device in devices:
            if device[self.PATH_IN_CONTAINER]:
                actual_devices.append(':'.join(
                    [device[self.PATH_ON_HOST],
                     device[self.PATH_IN_CONTAINER],
                     device[self.PERMISSIONS]]))
            else:
                actual_devices.append(':'.join(
                    [device[self.PATH_ON_HOST],
                     device[self.PATH_ON_HOST],
                     device[self.PERMISSIONS]]))
        return actual_devices

    def _get_container_status(self, container_id):
        client = self.get_client()
        info = client.inspect_container(container_id)
        return info['State']

    def check_create_complete(self, container_id):
        status = self._get_container_status(container_id)
        exit_status = status.get('ExitCode')
        if exit_status is not None and exit_status != 0:
            logs = self.get_client().logs(self.resource_id)
            raise exception.ResourceInError(resource_status=self.FAILED,
                                            status_reason=logs)
        return status['Running']

    def handle_delete(self):
        if self.resource_id is None:
            return
        client = self.get_client()
        try:
            client.kill(self.resource_id)
        except docker.errors.APIError as ex:
            if ex.response.status_code != 404:
                raise
        return self.resource_id

    def check_delete_complete(self, container_id):
        if container_id is None:
            return True
        try:
            status = self._get_container_status(container_id)
            if not status['Running']:
                client = self.get_client()
                client.remove_container(container_id)
        except docker.errors.APIError as ex:
            if ex.response.status_code == 404:
                return True
            raise

        return False

    def handle_suspend(self):
        if not self.resource_id:
            return
        client = self.get_client()
        client.stop(self.resource_id)
        return self.resource_id

    def check_suspend_complete(self, container_id):
        status = self._get_container_status(container_id)
        return (not status['Running'])

    def handle_resume(self):
        if not self.resource_id:
            return
        client = self.get_client()
        client.start(self.resource_id)
        return self.resource_id

    def check_resume_complete(self, container_id):
        status = self._get_container_status(container_id)
        return status['Running']

    def validate(self):
        super(DockerContainer, self).validate()
        self._validate_arg_for_api_version()

    def _validate_arg_for_api_version(self):
        version = None
        for key in MIN_API_VERSION_MAP:
            if self.properties[key]:
                if not version:
                    client = self.get_client()
                    version = client.version()['ApiVersion']
                min_version = MIN_API_VERSION_MAP[key]
                if compare_version(min_version, version) < 0:
                    raise InvalidArgForVersion(arg=key,
                                               min_version=min_version)


def resource_mapping():
    return {
        'DockerInc::Docker::Container': DockerContainer,
    }


def available_resource_mapping():
    if DOCKER_INSTALLED:
        return resource_mapping()
    else:
        LOG.warning("Docker plug-in loaded, but docker lib "
                    "not installed.")
        return {}


def compare_version(v1, v2):
    s1 = distutils.version.StrictVersion(v1)
    s2 = distutils.version.StrictVersion(v2)
    if s1 == s2:
        return 0
    elif s1 > s2:
        return -1
    else:
        return 1


class InvalidArgForVersion(exception.HeatException):
    msg_fmt = _('"%(arg)s" is not supported for API version '
                '< "%(min_version)s"')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import os
import subprocess
from unittest import mock


from heat.cloudinit import loguserdata
from heat.tests import common


class FakeCiVersion(object):
    def __init__(self, version):
        self.version = version


class LoguserdataTest(common.HeatTestCase):

    @mock.patch('pkg_resources.get_distribution')
    def test_ci_version_with_pkg_resources(self, mock_get):
        # Setup
        returned_versions = [
            FakeCiVersion('0.5.0'),
            FakeCiVersion('0.5.9'),
            FakeCiVersion('0.6.0'),
            FakeCiVersion('0.7.0'),
            FakeCiVersion('1.0'),
            FakeCiVersion('2.0'),
        ]
        mock_get.side_effect = returned_versions

        # Test & Verify
        self.assertFalse(loguserdata.chk_ci_version())
        self.assertFalse(loguserdata.chk_ci_version())
        self.assertTrue(loguserdata.chk_ci_version())
        self.assertTrue(loguserdata.chk_ci_version())
        self.assertTrue(loguserdata.chk_ci_version())
        self.assertTrue(loguserdata.chk_ci_version())
        self.assertEqual(6, mock_get.call_count)

    @mock.patch('pkg_resources.get_distribution')
    @mock.patch('subprocess.Popen')
    def test_ci_version_with_subprocess(self, mock_popen,
                                        mock_get_distribution):
        # Setup
        mock_get_distribution.side_effect = Exception()

        popen_return = [
            [None, 'cloud-init 0.0.5\n'],
            [None, 'cloud-init 0.7.5\n'],
        ]
        mock_popen.return_value = mock.MagicMock()
        mock_popen.return_value.communicate.side_effect = popen_return

        # Test & Verify
        self.assertFalse(loguserdata.chk_ci_version())
        self.assertTrue(loguserdata.chk_ci_version())
        self.assertEqual(2, mock_get_distribution.call_count)

    @mock.patch('pkg_resources.get_distribution')
    @mock.patch('subprocess.Popen')
    def test_ci_version_with_subprocess_exception(self, mock_popen,
                                                  mock_get_distribution):
        # Setup
        mock_get_distribution.side_effect = Exception()
        mock_popen.return_value = mock.MagicMock()
        mock_popen.return_value.communicate.return_value = ['non-empty',
                                                            'irrelevant']

        # Test
        self.assertRaises(Exception, loguserdata.chk_ci_version)  # noqa
        self.assertEqual(1, mock_get_distribution.call_count)

    @mock.patch('subprocess.Popen')
    def test_call(self, mock_popen):
        # Setup
        mock_popen.return_value = mock.MagicMock()
        mock_popen.return_value.communicate.return_value = ['a', 'b']
        mock_popen.return_value.returncode = 0

        # Test
        return_code = loguserdata.call(['foo', 'bar'])

        # Verify
        mock_popen.assert_called_once_with(['foo', 'bar'],
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE)
        self.assertEqual(0, return_code)

    @mock.patch('sys.exc_info')
    @mock.patch('subprocess.Popen')
    def test_call_oserror_enoexec(self, mock_popen, mock_exc_info):
        # Setup
        mock_popen.side_effect = OSError()
        no_exec = mock.MagicMock(errno=errno.ENOEXEC)
        mock_exc_info.return_value = None, no_exec, None

        # Test
        return_code = loguserdata.call(['foo', 'bar'])

        # Verify
        self.assertEqual(os.EX_OK, return_code)

    @mock.patch('sys.exc_info')
    @mock.patch('subprocess.Popen')
    def test_call_oserror_other(self, mock_popen, mock_exc_info):
        # Setup
        mock_popen.side_effect = OSError()
        no_exec = mock.MagicMock(errno='foo')
        mock_exc_info.return_value = None, no_exec, None

        # Test
        return_code = loguserdata.call(['foo', 'bar'])

        # Verify
        self.assertEqual(os.EX_OSERR, return_code)

    @mock.patch('sys.exc_info')
    @mock.patch('subprocess.Popen')
    def test_call_exception(self, mock_popen, mock_exc_info):
        # Setup
        mock_popen.side_effect = Exception()
        no_exec = mock.MagicMock(errno='irrelevant')
        mock_exc_info.return_value = None, no_exec, None

        # Test
        return_code = loguserdata.call(['foo', 'bar'])

        # Verify
        self.assertEqual(os.EX_SOFTWARE, return_code)

    @mock.patch('pkg_resources.get_distribution')
    @mock.patch('os.chmod')
    @mock.patch('heat.cloudinit.loguserdata.call')
    def test_main(self, mock_call, mock_chmod, mock_get):
        # Setup
        mock_get.return_value = FakeCiVersion('1.0')
        mock_call.return_value = 10

        # Test
        return_code = loguserdata.main()

        # Verify
        expected_path = os.path.join(loguserdata.VAR_PATH, 'cfn-userdata')
        mock_chmod.assert_called_once_with(expected_path, int('700', 8))
        self.assertEqual(10, return_code)

    @mock.patch('pkg_resources.get_distribution')
    def test_main_failed_ci_version(self, mock_get):
        # Setup
        mock_get.return_value = FakeCiVersion('0.0.0')

        # Test
        return_code = loguserdata.main()

        # Verify
        self.assertEqual(-1, return_code)



#
#    Copyright 2015 IBM Corp.
#
#    All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support
from heat.engine import translation


class Listener(neutron.NeutronResource):
    """A resource for managing LBaaS v2 Listeners.

    This resource creates and manages Neutron LBaaS v2 Listeners,
    which represent a listening endpoint for the vip.
    """

    support_status = support.SupportStatus(version='6.0.0')

    required_service_extension = 'lbaasv2'

    entity = 'listener'

    PROPERTIES = (
        PROTOCOL_PORT, PROTOCOL, LOADBALANCER, DEFAULT_POOL, NAME,
        ADMIN_STATE_UP, DESCRIPTION, DEFAULT_TLS_CONTAINER_REF,
        SNI_CONTAINER_REFS, CONNECTION_LIMIT, TENANT_ID
    ) = (
        'protocol_port', 'protocol', 'loadbalancer', 'default_pool', 'name',
        'admin_state_up', 'description', 'default_tls_container_ref',
        'sni_container_refs', 'connection_limit', 'tenant_id'
    )

    PROTOCOLS = (
        TCP, HTTP, HTTPS, TERMINATED_HTTPS,
    ) = (
        'TCP', 'HTTP', 'HTTPS', 'TERMINATED_HTTPS',
    )

    ATTRIBUTES = (
        LOADBALANCERS_ATTR, DEFAULT_POOL_ID_ATTR
    ) = (
        'loadbalancers', 'default_pool_id'
    )

    properties_schema = {
        PROTOCOL_PORT: properties.Schema(
            properties.Schema.INTEGER,
            _('TCP or UDP port on which to listen for client traffic.'),
            required=True,
            constraints=[
                constraints.Range(1, 65535),
            ]
        ),
        PROTOCOL: properties.Schema(
            properties.Schema.STRING,
            _('Protocol on which to listen for the client traffic.'),
            required=True,
            constraints=[
                constraints.AllowedValues(PROTOCOLS),
            ]
        ),
        LOADBALANCER: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of the load balancer with which listener '
              'is associated.'),
            constraints=[
                constraints.CustomConstraint('neutron.lbaas.loadbalancer')
            ]
        ),
        DEFAULT_POOL: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of the default pool for the listener. Requires '
              'shared_pools service extension.'),
            update_allowed=True,
            constraints=[
                constraints.CustomConstraint('neutron.lbaas.pool')
            ],
            support_status=support.SupportStatus(version='9.0.0')
        ),
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of this listener.'),
            update_allowed=True
        ),
        ADMIN_STATE_UP: properties.Schema(
            properties.Schema.BOOLEAN,
            _('The administrative state of this listener.'),
            update_allowed=True,
            default=True
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description of this listener.'),
            update_allowed=True,
            default=''
        ),
        DEFAULT_TLS_CONTAINER_REF: properties.Schema(
            properties.Schema.STRING,
            _('Default TLS container reference to retrieve TLS '
              'information.'),
            update_allowed=True
        ),
        SNI_CONTAINER_REFS: properties.Schema(
            properties.Schema.LIST,
            _('List of TLS container references for SNI.'),
            update_allowed=True
        ),
        CONNECTION_LIMIT: properties.Schema(
            properties.Schema.INTEGER,
            _('The maximum number of connections permitted for this '
              'load balancer. Defaults to -1, which is infinite.'),
            update_allowed=True,
            default=-1,
            constraints=[
                constraints.Range(min=-1),
            ]
        ),
        TENANT_ID: properties.Schema(
            properties.Schema.STRING,
            _('The ID of the tenant who owns the listener.')
        ),
    }

    attributes_schema = {
        LOADBALANCERS_ATTR: attributes.Schema(
            _('ID of the load balancer this listener is associated to.'),
            type=attributes.Schema.LIST
        ),
        DEFAULT_POOL_ID_ATTR: attributes.Schema(
            _('ID of the default pool this listener is associated to.'),
            type=attributes.Schema.STRING
        )
    }

    def __init__(self, name, definition, stack):
        super(Listener, self).__init__(name, definition, stack)
        self._lb_id = None

    def translation_rules(self, props):
        client_plugin = self.client_plugin()
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.LOADBALANCER],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_LOADBALANCER
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.DEFAULT_POOL],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_LB_POOL
            ),
        ]

    def validate(self):
        res = super(Listener, self).validate()
        if res:
            return res

        if (self.properties[self.LOADBALANCER] is None
                and self.properties[self.DEFAULT_POOL] is None):
            raise exception.PropertyUnspecifiedError(self.LOADBALANCER,
                                                     self.DEFAULT_POOL)

        if self.properties[self.PROTOCOL] == self.TERMINATED_HTTPS:
            if self.properties[self.DEFAULT_TLS_CONTAINER_REF] is None:
                msg = (_('Property %(ref)s required when protocol is '
                       '%(term)s.') % {'ref': self.DEFAULT_TLS_CONTAINER_REF,
                                       'term': self.TERMINATED_HTTPS})
                raise exception.StackValidationFailed(message=msg)

    @property
    def lb_id(self):
        if self._lb_id:
            return self._lb_id

        self._lb_id = self.properties[self.LOADBALANCER]
        if self._lb_id is None:
            pool_id = self.properties[self.DEFAULT_POOL]
            pool = self.client().show_pool(pool_id)['pool']
            self._lb_id = pool['loadbalancers'][0]['id']
        return self._lb_id

    def _check_lb_status(self):
        return self.client_plugin().check_lb_status(self.lb_id)

    def handle_create(self):
        properties = self.prepare_properties(
            self.properties,
            self.physical_resource_name())

        if self.LOADBALANCER in properties:
            properties['loadbalancer_id'] = properties.pop(self.LOADBALANCER)
        if self.DEFAULT_POOL in properties:
            properties['default_pool_id'] = properties.pop(self.DEFAULT_POOL)
        return properties

    def check_create_complete(self, properties):
        if self.resource_id is None:
            try:
                listener = self.client().create_listener(
                    {'listener': properties})['listener']
                self.resource_id_set(listener['id'])
            except Exception as ex:
                if self.client_plugin().is_invalid(ex):
                    return False
                raise

        return self._check_lb_status()

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        self._update_called = False
        self.properties = json_snippet.properties(self.properties_schema,
                                                  self.context)
        return prop_diff

    def check_update_complete(self, prop_diff):
        if not prop_diff:
            return True

        if self.DEFAULT_POOL in prop_diff:
            prop_diff['default_pool_id'] = prop_diff.pop(self.DEFAULT_POOL)
        if not self._update_called:
            try:
                self.client().update_listener(self.resource_id,
                                              {'listener': prop_diff})
                self._update_called = True
            except Exception as ex:
                if self.client_plugin().is_invalid(ex):
                    return False
                raise

        return self._check_lb_status()

    def handle_delete(self):
        self._delete_called = False

    def check_delete_complete(self, data):
        if self.resource_id is None:
            return True

        if not self._delete_called:
            try:
                self.client().delete_listener(self.resource_id)
                self._delete_called = True
            except Exception as ex:
                if self.client_plugin().is_invalid(ex):
                    return False
                elif self.client_plugin().is_not_found(ex):
                    return True
                raise

        return self._check_lb_status()


def resource_mapping():
    return {
        'OS::Neutron::LBaaS::Listener': Listener,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support


class GenericContainer(resource.Resource):
    """A resource for creating Barbican generic container.

    A generic container is used for any type of secret that a user
    may wish to aggregate. There are no restrictions on the amount
    of secrets that can be held within this container.
    """

    support_status = support.SupportStatus(version='6.0.0')

    default_client_name = 'barbican'

    entity = 'containers'

    PROPERTIES = (
        NAME, SECRETS,
    ) = (
        'name', 'secrets',
    )

    ATTRIBUTES = (
        STATUS, CONTAINER_REF, SECRET_REFS, CONSUMERS,
    ) = (
        'status', 'container_ref', 'secret_refs', 'consumers',
    )

    _SECRETS_PROPERTIES = (
        NAME, REF,
    ) = (
        'name', 'ref'
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Human-readable name for the container.'),
        ),
        SECRETS: properties.Schema(
            properties.Schema.LIST,
            _('References to secrets that will be stored in container.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    NAME: properties.Schema(
                        properties.Schema.STRING,
                        _('Name of the secret.'),
                        required=True
                    ),
                    REF: properties.Schema(
                        properties.Schema.STRING,
                        _('Reference to the secret.'),
                        required=True,
                        constraints=[constraints.CustomConstraint(
                            'barbican.secret')],
                    ),
                }
            ),
        ),
    }

    attributes_schema = {
        STATUS: attributes.Schema(
            _('The status of the container.'),
            type=attributes.Schema.STRING
        ),
        CONTAINER_REF: attributes.Schema(
            _('The URI to the container.'),
            type=attributes.Schema.STRING
        ),
        SECRET_REFS: attributes.Schema(
            _('The URIs to secrets stored in container.'),
            type=attributes.Schema.MAP
        ),
        CONSUMERS: attributes.Schema(
            _('The URIs to container consumers.'),
            type=attributes.Schema.LIST
        ),
    }

    def get_refs(self):
        secrets = self.properties.get(self.SECRETS) or []
        return [secret['ref'] for secret in secrets]

    def validate(self):
        super(GenericContainer, self).validate()
        refs = self.get_refs()
        if len(refs) != len(set(refs)):
            msg = _("Duplicate refs are not allowed.")
            raise exception.StackValidationFailed(message=msg)

    def create_container(self):
        if self.properties[self.SECRETS]:
            secrets = dict((secret['name'], secret['ref'])
                           for secret in self.properties[self.SECRETS])
        else:
            secrets = {}
        info = {'secret_refs': secrets}
        if self.properties[self.NAME] is not None:
            info.update({'name': self.properties[self.NAME]})
        return self.client_plugin().create_generic_container(**info)

    def handle_create(self):
        container_ref = self.create_container().store()
        self.resource_id_set(container_ref)
        return container_ref

    def check_create_complete(self, container_href):
        container = self.client().containers.get(container_href)

        if container.status == 'ERROR':
            reason = container.error_reason
            code = container.error_status_code
            msg = (_("Container '%(name)s' creation failed: "
                     "%(code)s - %(reason)s")
                   % {'name': self.name, 'code': code, 'reason': reason})
            raise exception.ResourceInError(
                status_reason=msg, resource_status=container.status)

        return container.status == 'ACTIVE'

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        container = self.client().containers.get(self.resource_id)
        return getattr(container, name, None)


class CertificateContainer(GenericContainer):
    """A resource for creating barbican certificate container.

    A certificate container is used for storing the secrets that
    are relevant to certificates.
    """

    PROPERTIES = (
        NAME, CERTIFICATE_REF, PRIVATE_KEY_REF,
        PRIVATE_KEY_PASSPHRASE_REF, INTERMEDIATES_REF,
    ) = (
        'name', 'certificate_ref', 'private_key_ref',
        'private_key_passphrase_ref', 'intermediates_ref',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Human-readable name for the container.'),
        ),
        CERTIFICATE_REF: properties.Schema(
            properties.Schema.STRING,
            _('Reference to certificate.'),
            constraints=[constraints.CustomConstraint('barbican.secret')],
        ),
        PRIVATE_KEY_REF: properties.Schema(
            properties.Schema.STRING,
            _('Reference to private key.'),
            constraints=[constraints.CustomConstraint('barbican.secret')],
        ),
        PRIVATE_KEY_PASSPHRASE_REF: properties.Schema(
            properties.Schema.STRING,
            _('Reference to private key passphrase.'),
            constraints=[constraints.CustomConstraint('barbican.secret')],
        ),
        INTERMEDIATES_REF: properties.Schema(
            properties.Schema.STRING,
            _('Reference to intermediates.'),
            constraints=[constraints.CustomConstraint('barbican.secret')],
        ),
    }

    def create_container(self):
        info = dict((k, v) for k, v in self.properties.items()
                    if v is not None)
        return self.client_plugin().create_certificate(**info)

    def get_refs(self):
        return [v for k, v in self.properties.items()
                if (k != self.NAME and v is not None)]


class RSAContainer(GenericContainer):
    """A resource for creating barbican RSA container.

    An RSA container is used for storing RSA public keys, private keys,
    and private key pass phrases.
    """

    PROPERTIES = (
        NAME, PRIVATE_KEY_REF, PRIVATE_KEY_PASSPHRASE_REF,
        PUBLIC_KEY_REF,
    ) = (
        'name', 'private_key_ref', 'private_key_passphrase_ref',
        'public_key_ref',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Human-readable name for the container.'),
        ),
        PRIVATE_KEY_REF: properties.Schema(
            properties.Schema.STRING,
            _('Reference to private key.'),
            constraints=[constraints.CustomConstraint('barbican.secret')],
        ),
        PRIVATE_KEY_PASSPHRASE_REF: properties.Schema(
            properties.Schema.STRING,
            _('Reference to private key passphrase.'),
            constraints=[constraints.CustomConstraint('barbican.secret')],
        ),
        PUBLIC_KEY_REF: properties.Schema(
            properties.Schema.STRING,
            _('Reference to public key.'),
            constraints=[constraints.CustomConstraint('barbican.secret')],
        ),
    }

    def create_container(self):
        info = dict((k, v) for k, v in self.properties.items()
                    if v is not None)
        return self.client_plugin().create_rsa(**info)

    def get_refs(self):
        return [v for k, v in self.properties.items()
                if (k != self.NAME and v is not None)]


def resource_mapping():
    return {
        'OS::Barbican::GenericContainer': GenericContainer,
        'OS::Barbican::CertificateContainer': CertificateContainer,
        'OS::Barbican::RSAContainer': RSAContainer
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
from urllib import parse as urlparse

from heat.api.openstack.v1.views import views_common
from heat.tests import common


class TestViewsCommon(common.HeatTestCase):
    def setUp(self):
        super(TestViewsCommon, self).setUp()
        self.request = mock.Mock()
        self.stack1 = {
            'id': 'id1',
        }
        self.stack2 = {
            'id': 'id2',
        }

    def setUpGetCollectionLinks(self):
        self.items = [self.stack1, self.stack2]
        self.request.params = {'limit': '2'}
        self.request.path_url = "http://example.com/fake/path"

    def test_get_collection_links_creates_next(self):
        self.setUpGetCollectionLinks()
        links = views_common.get_collection_links(self.request, self.items)

        expected_params = {'marker': ['id2'], 'limit': ['2']}
        next_link = list(filter(
            lambda link: link['rel'] == 'next', links)).pop()
        self.assertEqual('next', next_link['rel'])
        url_path, url_params = next_link['href'].split('?', 1)
        self.assertEqual(url_path, self.request.path_url)
        self.assertEqual(expected_params, urlparse.parse_qs(url_params))

    def test_get_collection_links_doesnt_create_next_if_no_limit(self):
        self.setUpGetCollectionLinks()
        del self.request.params['limit']
        links = views_common.get_collection_links(self.request, self.items)

        self.assertEqual([], links)

    def test_get_collection_links_doesnt_create_next_if_page_not_full(self):
        self.setUpGetCollectionLinks()
        self.request.params['limit'] = '10'
        links = views_common.get_collection_links(self.request, self.items)

        self.assertEqual([], links)

    def test_get_collection_links_overwrites_url_marker(self):
        self.setUpGetCollectionLinks()
        self.request.params = {'limit': '2', 'marker': 'some_marker'}
        links = views_common.get_collection_links(self.request, self.items)
        expected_params = {'marker': ['id2'], 'limit': ['2']}
        next_link = list(filter(
            lambda link: link['rel'] == 'next', links)).pop()
        self.assertEqual('next', next_link['rel'])
        url_path, url_params = next_link['href'].split('?', 1)
        self.assertEqual(url_path, self.request.path_url)
        self.assertEqual(expected_params, urlparse.parse_qs(url_params))

    def test_get_collection_links_does_not_overwrite_other_params(self):
        self.setUpGetCollectionLinks()
        self.request.params = {'limit': '2', 'foo': 'bar'}
        links = views_common.get_collection_links(self.request, self.items)

        next_link = list(
            filter(lambda link: link['rel'] == 'next', links)).pop()
        url = next_link['href']
        query_string = urlparse.urlparse(url).query
        params = {}
        params.update(urlparse.parse_qsl(query_string))
        self.assertEqual('2', params['limit'])
        self.assertEqual('bar', params['foo'])

    def test_get_collection_links_handles_invalid_limits(self):
        self.setUpGetCollectionLinks()
        self.request.params = {'limit': 'foo'}
        links = views_common.get_collection_links(self.request, self.items)
        self.assertEqual([], links)

        self.request.params = {'limit': None}
        links = views_common.get_collection_links(self.request, self.items)
        self.assertEqual([], links)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support


class SecurityGroup(neutron.NeutronResource):
    """A resource for managing Neutron security groups.

    Security groups are sets of IP filter rules that are applied to an
    instance's networking. They are project specific, and project members can
    edit the default rules for their group and add new rules sets. All projects
    have a "default" security group, which is applied to instances that have no
    other security group defined.
    """

    required_service_extension = 'security-group'

    entity = 'security_group'

    support_status = support.SupportStatus(version='2014.1')

    PROPERTIES = (
        NAME, DESCRIPTION, RULES,
    ) = (
        'name', 'description', 'rules',
    )

    _RULE_KEYS = (
        RULE_DIRECTION, RULE_ETHERTYPE, RULE_PORT_RANGE_MIN,
        RULE_PORT_RANGE_MAX, RULE_PROTOCOL, RULE_REMOTE_MODE,
        RULE_REMOTE_GROUP_ID, RULE_REMOTE_IP_PREFIX,
    ) = (
        'direction', 'ethertype', 'port_range_min',
        'port_range_max', 'protocol', 'remote_mode',
        'remote_group_id', 'remote_ip_prefix',
    )

    _rule_schema = {
        RULE_DIRECTION: properties.Schema(
            properties.Schema.STRING,
            _('The direction in which the security group rule is applied. '
              'For a compute instance, an ingress security group rule '
              'matches traffic that is incoming (ingress) for that '
              'instance. An egress rule is applied to traffic leaving '
              'the instance.'),
            default='ingress',
            constraints=[
                constraints.AllowedValues(['ingress', 'egress']),
            ]
        ),
        RULE_ETHERTYPE: properties.Schema(
            properties.Schema.STRING,
            _('Ethertype of the traffic.'),
            default='IPv4',
            constraints=[
                constraints.AllowedValues(['IPv4', 'IPv6']),
            ]
        ),
        RULE_PORT_RANGE_MIN: properties.Schema(
            properties.Schema.INTEGER,
            _('The minimum port number in the range that is matched by the '
              'security group rule. If the protocol is TCP or UDP, this '
              'value must be less than or equal to the value of the '
              'port_range_max attribute. If the protocol is ICMP, this '
              'value must be an ICMP type.'),
            constraints=[
                constraints.Range(0, 65535)
            ]
        ),
        RULE_PORT_RANGE_MAX: properties.Schema(
            properties.Schema.INTEGER,
            _('The maximum port number in the range that is matched by the '
              'security group rule. The port_range_min attribute constrains '
              'the port_range_max attribute. If the protocol is ICMP, this '
              'value must be an ICMP type.'),
            constraints=[
                constraints.Range(0, 65535)
            ]
        ),
        RULE_PROTOCOL: properties.Schema(
            properties.Schema.STRING,
            _('The protocol that is matched by the security group rule. '
              'Valid values include tcp, udp, and icmp.')
        ),
        RULE_REMOTE_MODE: properties.Schema(
            properties.Schema.STRING,
            _('Whether to specify a remote group or a remote IP prefix.'),
            default='remote_ip_prefix',
            constraints=[
                constraints.AllowedValues(['remote_ip_prefix',
                                           'remote_group_id']),
            ]
        ),
        RULE_REMOTE_GROUP_ID: properties.Schema(
            properties.Schema.STRING,
            _('The remote group ID to be associated with this security group '
              'rule. If no value is specified then this rule will use this '
              'security group for the remote_group_id. The remote mode '
              'parameter must be set to "remote_group_id".'),
            constraints=[
                constraints.CustomConstraint('neutron.security_group')
            ]
        ),
        RULE_REMOTE_IP_PREFIX: properties.Schema(
            properties.Schema.STRING,
            _('The remote IP prefix (CIDR) to be associated with this '
              'security group rule.'),
            constraints=[
                constraints.CustomConstraint('net_cidr')
            ]
        ),
    }

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('A string specifying a symbolic name for the security group, '
              'which is not required to be unique.'),
            update_allowed=True
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description of the security group.'),
            update_allowed=True
        ),
        RULES: properties.Schema(
            properties.Schema.LIST,
            _('List of security group rules.'),
            default=[],
            schema=properties.Schema(
                properties.Schema.MAP,
                schema=_rule_schema
            ),
            update_allowed=True
        ),
    }

    default_egress_rules = [
        {"direction": "egress", "ethertype": "IPv4"},
        {"direction": "egress", "ethertype": "IPv6"}
    ]

    def validate(self):
        super(SecurityGroup, self).validate()
        if self.properties[self.NAME] == 'default':
            msg = _('Security groups cannot be assigned the name "default".')
            raise exception.StackValidationFailed(message=msg)

    def handle_create(self):
        props = self.prepare_properties(
            self.properties,
            self.physical_resource_name())
        rules = props.pop(self.RULES, [])

        sec = self.client().create_security_group(
            {'security_group': props})['security_group']

        self.resource_id_set(sec['id'])
        self._create_rules(rules)

    def _format_rule(self, r):
        rule = dict(r)
        rule['security_group_id'] = self.resource_id

        if 'remote_mode' in rule:
            remote_mode = rule.get(self.RULE_REMOTE_MODE)
            del(rule[self.RULE_REMOTE_MODE])

            if remote_mode == self.RULE_REMOTE_GROUP_ID:
                rule[self.RULE_REMOTE_IP_PREFIX] = None
                if not rule.get(self.RULE_REMOTE_GROUP_ID):
                    # if remote group is not specified then make this
                    # a self-referencing rule
                    rule[self.RULE_REMOTE_GROUP_ID] = self.resource_id
            else:
                rule[self.RULE_REMOTE_GROUP_ID] = None

        for key in (self.RULE_PORT_RANGE_MIN, self.RULE_PORT_RANGE_MAX):
            if rule.get(key) is not None:
                rule[key] = str(rule[key])
        return rule

    def _create_rules(self, rules):
        egress_deleted = False

        for i in rules:
            if i[self.RULE_DIRECTION] == 'egress' and not egress_deleted:
                # There is at least one egress rule, so delete the default
                # rules which allow all egress traffic
                egress_deleted = True

                def is_egress(rule):
                    return rule[self.RULE_DIRECTION] == 'egress'

                self._delete_rules(is_egress)

            rule = self._format_rule(i)

            try:
                self.client().create_security_group_rule(
                    {'security_group_rule': rule})
            except Exception as ex:
                if not self.client_plugin().is_conflict(ex):
                    raise

    def _delete_rules(self, to_delete=None):
        try:
            sec = self.client().show_security_group(
                self.resource_id)['security_group']
        except Exception as ex:
            self.client_plugin().ignore_not_found(ex)
        else:
            for rule in sec['security_group_rules']:
                if to_delete is None or to_delete(rule):
                    with self.client_plugin().ignore_not_found:
                        self.client().delete_security_group_rule(rule['id'])

    def handle_delete(self):
        if self.resource_id is None:
            return

        self._delete_rules()
        with self.client_plugin().ignore_not_found:
            self.client().delete_security_group(self.resource_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        # handle rules changes by:
        # * deleting all rules
        # * restoring the default egress rules
        # * creating new rules
        rules = None
        if self.RULES in prop_diff:
            rules = prop_diff.pop(self.RULES)
            self._delete_rules()
            self._create_rules(self.default_egress_rules)

        if prop_diff:
            self.prepare_update_properties(prop_diff)
            self.client().update_security_group(
                self.resource_id, {'security_group': prop_diff})
        if rules:
            self._create_rules(rules)


def resource_mapping():
    return {
        'OS::Neutron::SecurityGroup': SecurityGroup,
    }
